"""Generate observation list files based on default values and APT output files.

Authors
-------
    - Lauren Chambers
    - Johannes Sahlmann

Use
---
    ::
        from mirage.yaml import generate_observationlist
        generate_observationlist.get_observation_dict(xml_file, yaml_file, catalogs,
            parameter_defaults=None, verbose=False):

TODO
----
    - Determine solution to set default parameters explicitly in configuration file or similar.
    - Clarify use and role of FilterConfig

"""
import collections
import copy
import os

from astropy.table import Table, vstack
import numpy as np

from ..apt import read_apt_xml


# Get the mapping between the user-input catalog dictionary keys and
# the keys used in the full table of observation parameters
CAT_TYPE_MAPPING = {'point_source': 'PointsourceCatalog', 'galaxy': 'GalaxyCatalog',
                    'extended': 'ExtendedCatalog', 'moving_pointsource': 'MovingTargetList',
                    'moving_sersic': 'MovingTargetSersic', 'moving_extended': 'MovingTargetExtended',
                    'moving_target_to_track': 'MovingTargetToTrack',
                    'tso_imaging_catalog': 'ImagingTSOCatalog',
                    'tso_grism_catalog': 'GrismTSOCatalog'}

POSSIBLE_CATS = list(CAT_TYPE_MAPPING.keys())


def catalog_dictionary_per_observation(cats, obs_nums, targets, defaults):
    """Translate a dictionary of catalogs from a case of either:

    1. Separate catalogs for each target name
    2. Separate catalogs for each target name and instrument

    into a dictionary of catalogs for each instrument and observation

    Parameters
    ----------
    cats : dict
        Dictionary of catalogs. Can be:

        Same catalogs for all instruments within each observation
        catalogs = {'my_targ_1': {'point_source': 'ptsrc1.cat',
                                  'galaxy': 'galaxy1.cat',
                                  'extended': 'ex1.cat'},
                    'my_targ_2': {'point_source': 'ptsrc2.cat',
                                  'galaxy': 'galaxy2.cat',
                                  'extended': 'ex2.cat'}}

        Different catalogs for each instrument in each observation
        catalogs = {'my_targ_1': {'nircam': {'point_source': 'ptsrc1.cat',
                                             'galaxy': 'galaxy1.cat',
                                             'extended': 'ex1.cat'},
                                  'niriss': {'pointsource': 'ptsrc_nis.cat',
                                             'galaxy': 'galaxy_nis.cat'}},
                    'my_targ_2': {'nircam': {'point_source': 'ptsrc2.cat',
                                             'galaxy': 'galaxy2.cat',
                                             'extended': 'ex2.cat'}}}

    obs_nums : numpy.ndarray
        1D array of observation ID numbers

    targets : numpy.ndarray
        1d array of target names, with a 1:1 correspondence to obs_nums

    defaults : dict
        Dictionary of default catalog values

    Returns
    -------
    obs_cats : dict
        Dictionary of catalogs per observation, with keys that match
        those in the defaults

        obs_cats = {'001': {'nircam': {'PointsourceCatalog': 'ptsrc1.cat',
                                       'GalaxyCatalog': 'galaxy1.cat',
                                       'ExtendedCatalog': 'ex1.cat'},
                            'niriss': {'PointsourceCatalog': 'ptsrc_nis.cat',
                                       'GalaxyCatalog': 'galaxy_nis.cat'},
                            }
                    '002': {'nircam': {'PointsourceCatalog': 'ptsrc2.cat',
                                       'GalaxyCatalog': 'galaxy2.cat',
                                       'ExtendedCatalog': 'ex2.cat'},
                            'niriss': {'PointsourceCatalog': 'ptsrc_nis2.cat',
                                       'GalaxyCatalog': 'galaxy_nis2.cat'}
                            }
                   }

    """
    # Set up the output dictionary. Populate with keys for all observations
    # and default catalog values to cover any entries in obs_cats that are
    # note present
    obs_cats = {}
    for number in obs_nums:
        obs_cats[number] = {'nircam': {}, 'niriss': {}, 'fgs': {}}
        for cat_type in POSSIBLE_CATS:
            obs_cats[number]['nircam'][CAT_TYPE_MAPPING[cat_type]] = defaults[CAT_TYPE_MAPPING[cat_type]]
            obs_cats[number]['niriss'][CAT_TYPE_MAPPING[cat_type]] = defaults[CAT_TYPE_MAPPING[cat_type]]
            obs_cats[number]['fgs'][CAT_TYPE_MAPPING[cat_type]] = defaults[CAT_TYPE_MAPPING[cat_type]]

    # Loop over the keys in the top level of the input dictionary
    for key1 in cats:

        # Find the observation numbers that use this target
        match = np.array(targets) == key1

        # Check to see if the second level of the input dictionary is
        # a dictionary of catalogs, or a dictionary of instruments
        keys2 = cats[key1].keys()
        keys_present = [True if poss in keys2 else False for poss in POSSIBLE_CATS]
        if any(keys_present):
            # Dictionary contains catalog names, so we use the same catalogs
            # for all instruments

            # Loop over the observation numbers that use this target and
            # populate the entries for each with the catalog names. In
            # this case the catalog names are the same for all instruments
            for obs_number in obs_nums[match]:
                for key2 in keys2:
                    obs_cats[obs_number]['nircam'][CAT_TYPE_MAPPING[key2]] = cats[key1][key2]
                    obs_cats[obs_number]['niriss'][CAT_TYPE_MAPPING[key2]] = cats[key1][key2]
                    obs_cats[obs_number]['fgs'][CAT_TYPE_MAPPING[key2]] = cats[key1][key2]
        else:
            # Dictionary contains instrument names
            # Loop over observation numbers that use this target and
            # populate the different catalogs for each instrument
            for obs_number in obs_nums[match]:
                for instrument in keys2:
                    ctypes = cats[key1][instrument].keys()
                    for ctype in ctypes:
                        obs_cats[obs_number][instrument][CAT_TYPE_MAPPING[ctype]] = cats[key1][instrument][ctype]
    return obs_cats


def convert_background_dict(bkgd):
    """Given a dictionary of background rates where the keys are observation
    numbers and the values are strings or numbers, expand the dictionary to
    contain entries for each instrument, and channels in the case of nircam.
    This function primarily makes sense as a way to enable users to keep a
    single background level 'high', 'medium', 'low' that applies to all
    insturuments in a given observation.

    Parameters
    ----------
    bkgd : dict
        For example:
        background = {'001': 'high', '002': 'medium', '003': 2.3}

    Return
    ------
    new_bkgd : dict
        For example:
        background = {'001': {'nircam': {'sw': high, 'lw': high}, 'niriss': high, 'fgs': high},
                      '002': {'nircam': {'sw': medium, 'lw': medium}, 'niriss': medium, 'fgs': medium},
                      '003': {'nircam': {'sw': 2.3, 'lw': 2.3}, 'niriss': 2.3, 'fgs': 2.3}
    """
    new_bkgd = {}
    for key, value in bkgd.items():
        new_bkgd[key] = {'nircam': {'sw': value, 'lw': value}, 'niriss': value, 'fgs': value}
    return new_bkgd


def dictionary_slice(dictionary, index):
    """Return a dictionary with only the i'th element from every list stored in a key.

    Parameters
    ----------
    dictionary
    index

    Returns
    -------

    """
    new_dict = {}
    for key in dictionary.keys():
        new_dict[key] = [dictionary[key][index]]
    return new_dict


def ensure_lower_case_background_keys(dictionary):
    """Ensure that the dictionary keys in the nested dictionary are all
    lower case. This was designed to be used on the user-input
    background level dictionary

    Parameters
    ----------
    dictionary : dict
        Nested dictionary of background values
        background = {'001': {'nircam': {'sw': 0.2, 'lw': 0.3}, 'niriss': 0.4, 'fgs': 0.2},
                      '002': {'nircam': {'sw': 'medium', 'lw': 'high'}, 'niriss': 'low', 'fgs': 'high'},
                      '003': {'nircam': {'sw': 0.75, 'lw': 'high'}, 'niriss': 0.2, 'fgs': 0.1}}

    Returns
    -------
    new_dict : dict
        Same as input dictionary, but with all keys converted to lower case
    """
    new_dict = {}
    for observation_number, obs_dict in dictionary.items():
        new_dict[observation_number] = {}
        for instrument, inst_data in obs_dict.items():
            # nircam with it's dictionary of sw and lw doesn't necessarily
            # have to be present, if the input proposal has no nircam
            # observations
            if isinstance(inst_data, collections.Mapping):
                new_dict[observation_number][instrument.lower()] = {}
                for channel, channel_val in inst_data.items():
                    new_dict[observation_number][instrument.lower()][channel.lower()] = channel_val
            else:
                new_dict[observation_number][instrument.lower()] = inst_data
    return new_dict


def ensure_lower_case_keys(dictionary):
    """Ensure that the dictionary keys in the nested dictionary are all
    lower case. This was designed to be used on the user-input
    background level dictionary

    Parameters
    ----------
    dictionary : dict
        Nested dictionary of background values
        background = {'001': {'nircam': {'sw': 0.2, 'lw': 0.3}, 'niriss': 0.4, 'fgs': 0.2},
                      '002': {'nircam': {'sw': 'medium', 'lw': 'high'}, 'niriss': 'low', 'fgs': 'high'},
                      '003': {'nircam': {'sw': 0.75, 'lw': 'high'}, 'niriss': 0.2, 'fgs': 0.1}}

    Returns
    -------
    new_dict : dict
        Same as input dictionary, but with all keys converted to lower case
    """
    new_dict = {}
    for key1, value1 in dictionary.items():
        new_dict[key1] = {}
        for key2, value2 in value1.items():
            # nircam with it's dictionary of sw and lw doesn't necessarily
            # have to be present, if the input proposal has no nircam
            # observations
            if isinstance(value2, collections.Mapping):
                new_dict[key1][key2.lower()] = {}
                for key3, value3 in value2.items():
                    new_dict[key1][key2.lower()][key3.lower()] = value3
            else:
                new_dict[key1][key2.lower()] = value2
    return new_dict


def ensure_lower_case_catalogs_keys(dictionary):
    """Ensure that the dictionary keys in the nested dictionary are all
    lower case. This was designed to be used on the user-input
    background level dictionary

    Parameters
    ----------
    dictionary : dict
        Nested dictionary of background values
        background = {'001': {'nircam': {'sw': 0.2, 'lw': 0.3}, 'niriss': 0.4, 'fgs': 0.2},
                      '002': {'nircam': {'sw': 'medium', 'lw': 'high'}, 'niriss': 'low', 'fgs': 'high'},
                      '003': {'nircam': {'sw': 0.75, 'lw': 'high'}, 'niriss': 0.2, 'fgs': 0.1}}

        catalogs = {'TARG1': {'point_source': 'ptsrc1.cat',
                              'galaxy': 'galaxy1.cat',
                              'extended': 'ex1.cat',
                              'moving_pointsource': 'mt_ptsrc1.cat',
                              'moving_sersic': 'mt_gal_1.cat',
                              'moving_extended': 'mt_ext_1.cat',
                              'moving_target_to_track': 'mt_track_1.cat'
                              },
                    'TARG2': {'point_source': 'ptsrc2.cat',
                              'galaxy': 'galaxy2.cat',
                              'extended': 'ex2.cat',
                              'moving_pointsource': 'mt_ptsrc2.cat',
                              'moving_sersic': 'mt_gal_2.cat',
                              'moving_extended': 'mt_ext_2.cat',
                              'moving_target_to_track': 'mt_track_2.cat'
                              }
                    }

        # Different catalogs for each instrument in each observation
        catalogs = {'TARG1': {'nircam': {'point_source': 'ptsrc_nrc_1.cat',
                                         'galaxy': 'galaxy_nrc_1.cat',
                                         'extended': 'ex_nrc_1.cat',
                                         'moving_pointsource': 'mt_ptsrc_nrc_1.cat',
                                         'moving_sersic': 'mt_gal_nrc_1.cat',
                                         'moving_extended': 'mt_ext_nrc_1.cat',
                                         'moving_target_to_track': 'mt_track_nrc_1.cat'
                                         },
                              'niriss': {'point_source': 'ptsrc_nis_1.cat',
                                         'galaxy': 'galaxy_nis_1.cat',
                                         'extended': 'ex_nis_1.cat',
                                         'moving_pointsource': 'mt_ptsrc_nis_1.cat',
                                         'moving_sersic': 'mt_gal_nis_1.cat',
                                         'moving_extended': 'mt_ext_nis_1.cat',
                                         'moving_target_to_track': 'mt_track_nis_1.cat'
                                         }
                              },
                    'TARG2': {'nircam': {'point_source': 'ptsrc_nrc_2.cat',
                                         'galaxy': 'galaxy_nrc_2.cat',
                                         'extended': 'ex_nrc_2.cat',
                                         'moving_pointsource': 'mt_ptsrc_nrc_2.cat',
                                         'moving_sersic': 'mt_gal_nrc_2.cat',
                                         'moving_extended': 'mt_ext_nrc_2.cat',
                                         'moving_target_to_track': 'mt_track_nrc_2.cat'
                                         },
                              'niriss': {'point_source': 'ptsrc_nis_2.cat',
                                         'galaxy': 'galaxy_nis_2.cat',
                                         'extended': 'ex_nis_2.cat',
                                         'moving_pointsource': 'mt_ptsrc_nis_2.cat',
                                         'moving_sersic': 'mt_gal_nis_2.cat',
                                         'moving_extended': 'mt_ext_nis_2.cat',
                                         'moving_target_to_track': 'mt_track_nis_2.cat'
                                         }
                              },
                    }

    Returns
    -------
    new_dict : dict
        Same as input dictionary, but with all keys converted to lower case
    """
    new_dict = {}
    for target_name, targ_dict in dictionary.items():
        new_dict[target_name] = {}
        for key2, value2 in targ_dict.items():
            if isinstance(value2, collections.Mapping):
                new_dict[target_name][key2.lower()] = {}
                for key3, value3 in targ_dict[key2].items():
                    new_dict[target_name][key2.lower()][key3.lower()] = value3
            else:
                new_dict[target_name][key2.lower()] = value2
    return new_dict


def expand_for_dithers(indict, verbose=True):
    """Expand a given dictionary to create one entry for each dither.

    Supports parallel observations.

    Moved here and modified from apt_inputs.py

    Parameters
    ----------
    indict : dict
        dictionary of observations

    Returns
    -------
    expanded : dict
        Dictionary, expanded to include a separate entry for
        each dither
    """
    expanded = {}
    for key in indict:
        expanded[key] = []

    # use astropy table operations to expand dithers while maintaining parallels in sync
    # implementation assumes that only one instrument is used in parallel
    table = Table(indict)
    table['row'] = np.arange(len(table))
    expanded_table = None #copy.deepcopy(table)

    # complication here is to handle cases with unsupported instruments (MIRI, NIRSpec) in parallel
    for i, row in enumerate(table['row']):
        number_of_dithers = np.int(table['number_of_dithers'][i])
        expand_prime_dithers_only = False
        expand_parallel_dithers = False

        # skip over parallel observations because they are already accounted for
        if table['ParallelInstrument'][i]:
            continue

        try:
            if (table['CoordinatedParallel'][i] == 'true') and (not table['ParallelInstrument'][i]) \
                    and (table['ParallelInstrument'][i + 1]) and (table['Instrument'][i] != table['Instrument'][i + 1]):
                expand_parallel_dithers = True
            else:
                expand_prime_dithers_only = True

        except IndexError:  # last row in table is not a parallel
            expand_prime_dithers_only = True

        if (table['CoordinatedParallel'][i] == 'false'):
            expand_prime_dithers_only = True

        if expand_prime_dithers_only and expand_parallel_dithers:
            raise RuntimeError('Possible conflict found when expanding for dithers.')

        if expand_parallel_dithers:
            dither_table = table[i:i + 2]

            if (number_of_dithers > 1):
                #replicate parallel observation n times
                dither_table = vstack([dither_table]*number_of_dithers)

            if expanded_table is None:
                expanded_table = dither_table
            else:
                # if verbose:
                #     print('Parallel: Adding {:>3d} rows to table with {:>3d} rows'.format(len(dither_table), len(expanded_table)))
                expanded_table = vstack((expanded_table, dither_table))

        elif expand_prime_dithers_only:
            # add row multiplied by number of dithers
            dither_table = vstack([table[i]]*number_of_dithers)

            if expanded_table is None:
                expanded_table = dither_table
            else:
                # print('Prime:    Adding {:>3d} rows to table with {:>3d} rows'.format(len(dither_table),
                #                                                             len(expanded_table)))
                expanded_table = vstack((expanded_table, dither_table))

    # set number of dithers to 1 after expansion
    expanded_table['number_of_dithers'] = np.ones(len(expanded_table)).astype(np.int)

    # NIRCam cannot handle when PrimaryDithers=None
    for index, value in enumerate(expanded_table['PrimaryDithers']):
        if value == 'None':
            expanded_table['PrimaryDithers'][index] = expanded_table['number_of_dithers'][index]

    expanded = {}
    for key in expanded_table.colnames:
        expanded[key] = np.array(expanded_table[key]).tolist()

    if verbose:
        print('Number of entries before expanding dithers: {}'.format(len(table)))
        print('Number of entries after expanding dithers:  {}'.format(len(expanded_table)))

    if verbose:
        for obs_id in np.unique(expanded_table['ObservationID']):
            print('Expanded table for Observation {} has {} entries'.format(obs_id, len(np.where(expanded_table['ObservationID']==obs_id)[0])))
    return expanded


def get_observation_dict(xml_file, yaml_file, catalogs,
                         parameter_overrides={'cosmic_rays': None, 'background': None, 'roll_angle': None, 'dates': None},
                         verbose=False):
    """Write observation list file (required mirage input) on the basis of APT files.

    Parameters
    ----------
    xml_file : str
        path to APT .xml file
    yaml_file : str
        output_file
    catalogs : dict
        Dictionary of catalog files, one entry per instrument. For NIRCam the entry has to be a
        dictionary itself, e.g. catalogs['nircam']['lw'] = somefile
        If the user prvides a list of catalogs, that list has to have one entry per observation in
        the program, accounting for any instrument used.
    parameter_overrides : dict
        Dictionary of default parameter value, e.g. date, roll angle, ...

    Returns
    -------
    xml_dict : dict
        Expanded dictionary that holds exposure information

    TODO
    ----
        Read default values from configuration file

    """
    # Read in filters from APT .xml file
    readxml_obj = read_apt_xml.ReadAPTXML()

    xml_dict = readxml_obj.read_xml(xml_file, verbose=verbose)

    # if verbose:
    #print('Summary of observation dictionary:')
    #for key in xml_dict.keys():
    #    print('{:<25}: number of elements is {:>5}'.format(key, len(xml_dict[key])))

    # create an expanded dictionary that contains lists of parameters expanded for dithers
    xml_dict = expand_for_dithers(xml_dict, verbose=verbose)
    #print('Summary of observation dictionary after expanding for dithers:')
    #for key in xml_dict.keys():
    #    print('{:<25}: number of elements is {:>5}'.format(key, len(xml_dict[key])))
    return_dict = None

    # array of unique instrument names
    used_instruments = np.unique(xml_dict['Instrument'])
    all_observation_ids = xml_dict['ObservationID']
    unique_observation_ids = np.unique(all_observation_ids).tolist()

    # List of target names from the proposal
    all_targets = xml_dict['TargetID']

    # Set default values. These are overwritten if there is an appropriate
    # entry in parameter_defaults
    default_values = {}
    default_values['Date'] = '2021-10-04'
    default_values['PAV3'] = '0.'
    default_values['PointsourceCatalog'] = 'None'
    default_values['GalaxyCatalog'] = 'None'
    default_values['ExtendedCatalog'] = 'None'
    default_values['ExtendedScale'] = '1.0'
    default_values['ExtendedCenter'] = '1024,1024'
    default_values['MovingTargetList'] = 'None'
    default_values['MovingTargetSersic'] = 'None'
    default_values['MovingTargetExtended'] = 'None'
    default_values['MovingTargetConvolveExtended'] = 'True'
    default_values['MovingTargetToTrack'] = 'None'
    default_values['ImagingTSOCatalog'] = 'None'
    default_values['GrismTSOCatalog'] = 'None'
    default_values['BackgroundRate_sw'] = 'low'
    default_values['BackgroundRate_lw'] = 'low'
    default_values['BackgroundRate'] = 'low'
    default_values['CosmicRayLibrary'] = 'SUNMAX'
    default_values['CosmicRayScale'] = 1.0
    default_parameter_name_list = ['MovingTargetConvolveExtended', 'ExtendedScale', 'ExtendedCenter']

    # Cosmic rays
    # Can be:
    # cr = {'library': 'SUNMAX', 'scale': 1.0}
    # cr = {'001': {'library': 'SUNMAX', 'scale': 1.2}}
    cosmic_rays = parameter_overrides['cosmic_rays']
    # Case where one value of library and scale are to be used for
    # all observations
    if cosmic_rays is not None:
        if 'library' in cosmic_rays.keys():
            default_values['CosmicRayLibrary'] = cosmic_rays['library']
            default_values['CosmicRayScale'] = cosmic_rays['scale']
            # Now set cosmic_rays to None so that it won't be used when looping
            # over observations below
            cosmic_rays = None
        else:
            # Case where different values are given for different observations
            # Just use cosmic_rays below when looping over observations
            pass

    # Background levels
    # background = 'high'
    # background = 22.2
    # background = {'001': 'high', '002': 'medium', '003': 22.3}
    # background = {'001': {'nircam': {'sw': 0.2, 'lw':0.3}, 'niriss': 0.4, 'fgs': 0.2}}
    background = parameter_overrides['background']
    if background is not None:
        if isinstance(background, str) or isinstance(background, float) or isinstance(background, int):
            default_values['BackgroundRate_sw'] = background
            default_values['BackgroundRate_lw'] = background
            default_values['BackgroundRate'] = background
            # Now set background to None so that it won't be used when looping
            # over observations below
            background = None
        else:
            bkeys = list(background.keys())
            if (isinstance(background[bkeys[0]], str) or isinstance(background[bkeys[0]], float) or isinstance(background[bkeys[0]], int)):
                # Case where one background is specified for all instruments
                # in each observation
                background = convert_background_dict(background)
            else:
                # Case where the user inputs the full dictionary, with a
                # background value for each instrument and observation.
                # Force all dictionary keys to be lower case.
                background = ensure_lower_case_keys(background)

    # Dates
    # dates = '2019-5-25'
    # dates = {'001': '2019-05-25', '002': '2019-11-15'}
    dates = parameter_overrides['dates']
    if dates is not None:
        if isinstance(dates, str):
            default_values['Date'] = dates
            # Now set dates to None so that it won't be used when looping
            # over observations below
            dates = None
        else:
            # Just use dates below when looping over observations
            pass

    #for key in parameter_defaults.keys():
    #    if key in default_values.keys():
    #        default_values[key] = parameter_defaults[key]

    # Roll angle, aka PAV3
    # pav3 = 34.5
    # pav3 = {'001': 34.5, '002': 154.5}
    pav3 = parameter_overrides['roll_angle']
    if pav3 is not None:
        if isinstance(pav3, float) or isinstance(pav3, int):
            default_values['PAV3'] = pav3
            # Now set pav3 to None so that it won't be used when looping
            # over observations below
            pav3 = None
        else:
            # Just use pav3 below when looping over observations
            pass

    # Catalogs
    # In the easy case where the same catalogs are to be used,
    # just populate the default values
    if catalogs is not None:
        cat_keys = catalogs.keys()
        keys_present = [True if poss in cat_keys else False for poss in POSSIBLE_CATS]
        if any(keys_present):
            for cat_key in cat_keys:
                if cat_key == 'point_source':
                    default_values['PointsourceCatalog'] = catalogs[cat_key]
                if cat_key == 'galaxy':
                    default_values['GalaxyCatalog'] = catalogs[cat_key]
                if cat_key == 'extended':
                    default_values['ExtendedCatalog'] = catalogs[cat_key]
                if cat_key == 'moving_pointsource':
                    default_values['MovingTargetList'] = catalogs[cat_key]
                if cat_key == 'moving_sersic':
                    default_values['MovingTargetSersic'] = catalogs[cat_key]
                if cat_key == 'moving_extended':
                    default_values['MovingTargetExtended'] = catalogs[cat_key]
                if cat_key == 'moving_target_to_track':
                    default_values['MovingTargetToTrack'] = catalogs[cat_key]
                if cat_key == 'tso_imaging_catalog':
                    default_values['ImagingTSOCatalog'] = catalogs[cat_key]
                if cat_key == 'tso_grism_catalog':
                    default_values['GrismTSOCatalog'] = catalogs[cat_key]
            # Now that we have modified the default values, set catalogs to
            # None so that it is not accessed later
            catalogs_per_observation = None
        else:
            # If the catalog dictionary is more complex, specifying different
            # catalogs for each target, or different catalogs for each instrument
            # and target, then translate this dictionary into a dictionary of
            # catalogs for each observation and instrument
            catalogs = ensure_lower_case_keys(catalogs)
            catalogs_per_observation = catalog_dictionary_per_observation(catalogs,
                                                                          np.array(all_observation_ids),
                                                                          np.array(all_targets),
                                                                          default_values)
    else:
        catalogs_per_observation = None

    # assemble string that will constitute the yaml content
    text_out = ["# Observation list created by generate_observationlist.py\n\n"]

    text = ['']
    entry_number = 0  # running number for every entry in the observation list

    # Create an instrument-specific counter to be used with input catalogs
    all_instruments = np.unique(xml_dict['Instrument'])
    counter = {}
    for inst in all_instruments:
        counter[inst] = 0

    entry_numbers = []
    observation_numbers = np.unique(xml_dict['ObservationID'])
    for observation_index, observation_number in enumerate(observation_numbers):
        first_index = xml_dict['ObservationID'].index(observation_number)
        text += [
            "Observation{}:\n".format(observation_number),
            "  Name: '{}'\n".format(xml_dict['ObservationName'][first_index])
            ]
        observation_rows = np.where(np.array(xml_dict['ObservationID']) == observation_number)[0]
        instruments_in_observation = np.unique(np.array(xml_dict['Instrument'])[observation_rows])
        for index in observation_rows:
            number_of_dithers = np.int(xml_dict['number_of_dithers'][index])
            instrument = xml_dict['Instrument'][index]
            for dither_index in range(number_of_dithers):

                # Get the proper date value
                if dates is None:
                    date_value = default_values['Date']
                else:
                    try:
                        date_value = dates[observation_number]
                    except KeyError:
                        print(("\n\nERROR: No date value specified for Observation {} in date dictionary. "
                               "Quitting.\n\n".format(observation_number)))
                        raise KeyError

                # Get the proper PAV3 value
                if pav3 is None:
                    pav3_value = default_values['PAV3']
                else:
                    try:
                        pav3_value = pav3[observation_number]
                    except KeyError:
                        print(("\n\nERROR: No roll angle value specified for Observation {} in roll_angle "
                               "dictionary. Quitting.\n\n".format(observation_number)))
                        raise KeyError

                # Get the proper catalog values
                if catalogs_per_observation is None:
                    ptsrc_catalog_value = default_values['PointsourceCatalog']
                    galaxy_catalog_value = default_values['GalaxyCatalog']
                    extended_catalog_value = default_values['ExtendedCatalog']
                    mov_ptsrc_catalog_value = default_values['MovingTargetList']
                    mov_sersic_catalog_value = default_values['MovingTargetSersic']
                    mov_extended_catalog_value = default_values['MovingTargetExtended']
                    mov_tracked_catalog_value = default_values['MovingTargetToTrack']
                    im_tso_catalog_value = default_values['ImagingTSOCatalog']
                    gr_tso_catalog_value = default_values['GrismTSOCatalog']
                else:
                    try:
                        catalogs_to_use = catalogs_per_observation[observation_number][instrument.lower()]
                    except KeyError:
                        print(("\n\nERROR: Missing observation number or instrument entry in catalog "
                               "dictionary. Failed to find catalogs[{}][{}]\n\n".format(observation_number,
                                                                                        instrument.lower())))
                        raise KeyError
                    ptsrc_catalog_value = catalogs_to_use['PointsourceCatalog']
                    galaxy_catalog_value = catalogs_to_use['GalaxyCatalog']
                    extended_catalog_value = catalogs_to_use['ExtendedCatalog']
                    mov_ptsrc_catalog_value = catalogs_to_use['MovingTargetList']
                    mov_sersic_catalog_value = catalogs_to_use['MovingTargetSersic']
                    mov_extended_catalog_value = catalogs_to_use['MovingTargetExtended']
                    mov_tracked_catalog_value = catalogs_to_use['MovingTargetToTrack']
                    im_tso_catalog_value = catalogs_to_use['ImagingTSOCatalog']
                    gr_tso_catalog_value = catalogs_to_use['GrismTSOCatalog']

                # Get the proper cosmic ray values
                if cosmic_rays is None:
                    cr_library_value = default_values['CosmicRayLibrary']
                    cr_scale_value = default_values['CosmicRayScale']
                else:
                    try:
                        cr_library_value = cosmic_rays[observation_number]['library']
                        cr_scale_value = cosmic_rays[observation_number]['scale']
                    except KeyError:
                        print(("\n\nERROR: No cosmic ray library and/or scale value specified for "
                               "Observation {} in cosmic_ray dictionary. Quitting.\n\n"
                               .format(observation_number)))
                        raise KeyError

                text += [
                    "  EntryNumber{}:\n".format(entry_number),
                    "    Instrument: {}\n".format(instrument),
                    "    Date: {}\n".format(date_value),
                    "    PAV3: {}\n".format(pav3_value),
                    "    DitherIndex: {}\n".format(dither_index),
                    "    CosmicRayLibrary: {}\n".format(cr_library_value),
                    "    CosmicRayScale: {}\n".format(cr_scale_value),
                ]
                if return_dict is None:
                    return_dict = dictionary_slice(xml_dict, index)
                else:
                    return_dict = read_apt_xml.append_dictionary(return_dict, dictionary_slice(xml_dict, index))
                if instrument.lower() in ['nircam', 'wfsc']:

                    sw_filt = xml_dict['ShortFilter'][index]
                    lw_filt = xml_dict['LongFilter'][index]

                    # Get the proper background rate
                    if background is None:
                        background_sw_value = default_values['BackgroundRate_sw']
                        background_lw_value = default_values['BackgroundRate_lw']
                    else:
                        try:
                            background_sw_value = background[observation_number]['nircam']['sw']
                            background_lw_value = background[observation_number]['nircam']['lw']
                        except KeyError:
                            print(("\n\nERROR: Missing entry in the background dictionary for NIRCam SW and/or "
                                   "LW channels, observation number: {}\n\n".format(observation_number)))
                            raise KeyError

                    text += [
                        "    FilterConfig:\n",
                        "      SW:\n",
                        "        Filter: {}\n".format(sw_filt),
                        "        PointSourceCatalog: {}\n".format(ptsrc_catalog_value),
                        "        GalaxyCatalog: {}\n".format(galaxy_catalog_value),
                        "        ExtendedCatalog: {}\n".format(extended_catalog_value),
                        "        MovingTargetList: {}\n".format(mov_ptsrc_catalog_value),
                        "        MovingTargetSersic: {}\n".format(mov_sersic_catalog_value),
                        "        MovingTargetExtended: {}\n".format(mov_extended_catalog_value),
                        "        MovingTargetToTrack: {}\n".format(mov_tracked_catalog_value),
                        "        ImagingTSOCatalog: {}\n".format(im_tso_catalog_value),
                        "        GrismTSOCatalog: {}\n".format(gr_tso_catalog_value),
                        "        BackgroundRate: {}\n".format(background_sw_value),
                        ]

                    for key in default_parameter_name_list:
                        text += ["        {}: {}\n".format(key, default_values[key])]

                    text += [
                        "      LW:\n",
                        "        Filter: {}\n".format(lw_filt),
                        "        PointSourceCatalog: {}\n".format(ptsrc_catalog_value),
                        "        GalaxyCatalog: {}\n".format(galaxy_catalog_value),
                        "        ExtendedCatalog: {}\n".format(extended_catalog_value),
                        "        MovingTargetList: {}\n".format(mov_ptsrc_catalog_value),
                        "        MovingTargetSersic: {}\n".format(mov_sersic_catalog_value),
                        "        MovingTargetExtended: {}\n".format(mov_extended_catalog_value),
                        "        MovingTargetToTrack: {}\n".format(mov_tracked_catalog_value),
                        "        ImagingTSOCatalog: {}\n".format(im_tso_catalog_value),
                        "        GrismTSOCatalog: {}\n".format(gr_tso_catalog_value),
                        "        BackgroundRate: {}\n".format(background_lw_value),
                        ]

                    for key in default_parameter_name_list:
                        text += ["        {}: {}\n".format(key, default_values[key])]

                elif instrument.lower() in ['niriss', 'fgs', 'nirspec', 'miri']:
                    if (instrument.lower() == 'niriss') and (xml_dict['APTTemplate'][index] in ['NirissExternalCalibration']):
                        filter_wheel_value = xml_dict['FilterWheel'][index]
                        pupil_wheel_value = xml_dict['PupilWheel'][index]
                        text += [
                            "    FilterWheel: {}\n".format(filter_wheel_value),
                            "    PupilWheel: {}\n".format(pupil_wheel_value),
                            ]
                        if 'CLEAR' in filter_wheel_value:
                            filter_value = pupil_wheel_value
                        elif ('CLEAR' in pupil_wheel_value) or ('NRM' in pupil_wheel_value):
                            filter_value = filter_wheel_value
                    else:
                        filter_value = xml_dict['Filter'][index]

                    # Get the proper background rate
                    if background is None:
                        background_value = default_values['BackgroundRate']
                    else:
                        try:
                            background_value = background[observation_number][instrument.lower()]
                        except KeyError:
                            print(("\n\nERROR: Missing entry in the background dictionary for observation "
                                   "number: {}, instrument: {}\n\n".format(observation_number, instrument)))
                            raise KeyError

                    text += [
                        "    Filter: {}\n".format(filter_value),
                        "    PointSourceCatalog: {}\n".format(ptsrc_catalog_value),
                        "    GalaxyCatalog: {}\n".format(galaxy_catalog_value),
                        "    ExtendedCatalog: {}\n".format(extended_catalog_value),
                        "    MovingTargetList: {}\n".format(mov_ptsrc_catalog_value),
                        "    MovingTargetSersic: {}\n".format(mov_sersic_catalog_value),
                        "    MovingTargetExtended: {}\n".format(mov_extended_catalog_value),
                        "    MovingTargetToTrack: {}\n".format(mov_tracked_catalog_value),
                        "    ImagingTSOCatalog: {}\n".format(im_tso_catalog_value),
                        "    GrismTSOCatalog: {}\n".format(gr_tso_catalog_value),
                        "    BackgroundRate: {}\n".format(background_value),
                        ]

                    for key in default_parameter_name_list:
                        text += ["    {}: {}\n".format(key, default_values[key])]

                entry_numbers.append(entry_number)
                entry_number += 1

        # Update the catalog counters for the instruments used in this observation
        #for inst_name in instruments_in_observation:
        #    counter[inst_name] += 1

    text_out += text

    return_dict['entry_number'] = entry_numbers

    # If the directory to hold the observation file does not yet exist, create it
    obs_dir = os.path.dirname(yaml_file)
    if obs_dir is not '' and os.path.isdir(obs_dir) is False:
        try:
            os.mkdir(obs_dir)
        except OSError:
            print("Creation of the directory {} failed".format(obs_dir))
        else:
            print("Successfully created the directory {} to hold the observation list file.".format(obs_dir))

    f = open(yaml_file, 'w')
    for line in text_out:
        f.write(line)
    f.close()
    print('\nWrote {} observations and {} entries to {}'.format(len(observation_numbers), entry_number, yaml_file))

    return return_dict



#! /usr/bin/env python

'''
Module for preparing a given dark current exposure for
integration with a ramp of simulated sources created
using (e.g.) catalog_seed_image.py.

Authors:
--------

    - Bryan Hilbert, Kevin Volk

Use:
----

    This module can be imported as such:

    ::

        from mirage.dark.dark_prep import DarkPrep
        dark = DarkPrep()
        dark.paramfile = 'my_parameters.yaml'
        dark.prepare()
'''

import sys
import os
import argparse
from math import floor
from glob import glob

import yaml
import pkg_resources
import numpy as np
from astropy.io import fits, ascii

import mirage
from mirage.utils import read_fits, utils, siaf_interface
from mirage.utils.file_splitting import find_file_splits
from mirage.reference_files import crds_tools


# Allowed instruments
INST_LIST = ['nircam', 'niriss', 'fgs']


class DarkPrep():
    def __init__(self, offline=False):
        """Instantiate DarkPrep object.

        Parameters
        ----------
        offline : bool
            If True, the check for the existence of the MIRAGE_DATA
            directory is skipped. This is primarily for Travis testing
        """
        self.offline = offline

        # Locate the module files, so that we know where to look
        # for config subdirectory
        self.modpath = pkg_resources.resource_filename('mirage', '')

        # Get the location of the MIRAGE_DATA environment
        # variable, so we know where to look for darks, CR,
        # PSF files, etc later
        self.env_var = 'MIRAGE_DATA'
        datadir = utils.expand_environment_variable(self.env_var, offline=offline)

        # Check that CRDS-related environment variables are set correctly
        self.crds_datadir = crds_tools.env_variables()

    def check_params(self):
        """Check for acceptible values for the input parameters in the
        yaml file.
        """
        # Check instrument name
        if self.params['Inst']['instrument'].lower() not in INST_LIST:
            raise ValueError(("WARNING: {} instrument not implemented within "
                              "simulator".format(self.params['Inst']['instrument'])))

        # If user requests not to use the pipeline,
        # make sure the input is a linearized dark, and not
        # a raw dark
        if self.params['Inst']['use_JWST_pipeline'] is False:
            if self.params['Reffiles']['linearized_darkfile'].lower() == 'none':
                raise ValueError(("WARNING: You specified no use of the JWST pipeline, but "
                                  "have not provided a linearized dark file to use. Without the "
                                  "pipeline, a raw dark cannot be used."))

        # If no raw dark nor linearized dark is given, then there is no
        # input to work with.
        if (self.params['Reffiles']['linearized_darkfile'].lower() == 'none' and
            self.params['Reffiles']['dark'].lower() == 'none'):
            raise ValueError(("WARNING: No raw nor linearized dark file given. Unable to proceed."))

        # Make sure nframe, nskip, ngroup are all integers
        try:
            self.params['Readout']['nframe'] = int(self.params['Readout']['nframe'])
        except ValueError:
            print("WARNING: Input value of nframe is not an integer.")

        try:
            self.params['Readout']['nskip'] = int(self.params['Readout']['nskip'])
        except ValueError:
            print("WARNING: Input value of nskip is not an integer.")

        try:
            self.params['Readout']['ngroup'] = int(self.params['Readout']['ngroup'])
        except ValueError:
            print("WARNING: Input value of ngroup is not an integer.")

        try:
            self.params['Readout']['nint'] = int(self.params['Readout']['nint'])
        except ValueError:
            print("WARNING: Input value of nint is not an integer.")

        # Make sure that the requested number of groups is
        # less than or equal to the maximum allowed. If you're
        # continuing on with an unknown readout pattern (not
        # recommended) then assume a max of 10 groups.
        # For science operations, ngroup is going to be limited
        # to 10 for all readout patterns except for the DEEP
        # patterns, which can go to 20. (FOR FULL FRAME!!)
        # Subarray exposures can use more!!
        match = self.readpatterns['name'] == self.params['Readout']['readpatt'].upper()
        if sum(match) == 1:
            maxgroups = self.readpatterns['maxgroups'].data[match][0]
        if sum(match) == 0:
            print(("Unrecognized readout pattern {}. Assuming a maximum allowed number of groups of 10."
                   .format(self.params['Readout']['readpatt'])))
            maxgroups = 10

        if (self.params['Readout']['ngroup'] > maxgroups):
            print(("WARNING: {} is limited to a maximum of {} groups. Proceeding with ngroup = {}."
                   .format(self.params['Readout']['readpatt'], maxgroups, maxgroups)))
            self.params['Readout']['readpatt'] = maxgroups

        # Check for entries in the parameter file that are None or blank,
        # indicating the step should be skipped. Create a dictionary of steps
        # and populate with True or False
        self.runStep = {}
        # self.runStep['linearity'] = self.check_run_step(self.params['Reffiles']['linearity'])
        self.runStep['badpixmask'] = self.check_run_step(self.params['Reffiles']['badpixmask'])
        self.runStep['linearized_darkfile'] = self.check_run_step(self.params['Reffiles']['linearized_darkfile'])
        self.runStep['saturation_lin_limit'] = self.check_run_step(self.params['Reffiles']['saturation'])
        self.runStep['superbias'] = self.check_run_step(self.params['Reffiles']['superbias'])
        self.runStep['linearity'] = self.check_run_step(self.params['Reffiles']['linearity'])

        # If the pipeline is going to be used to create
        # the linearized dark current ramp, make sure the
        # specified configuration files are present.
        if not self.runStep['linearized_darkfile']:
            dqcheck = self.check_run_step(self.params['newRamp']['dq_configfile'])
            if not dqcheck:
                raise FileNotFoundError(("WARNING: DQ pipeline step configuration file not provided. "
                                         "This file is needed to run the pipeline."))
            satcheck = self.check_run_step(self.params['newRamp']['sat_configfile'])
            if not satcheck:
                raise FileNotFoundError(("WARNING: Saturation pipeline step configuration file not provided. "
                                         "This file is needed to run the pipeline."))
            sbcheck = self.check_run_step(self.params['newRamp']['superbias_configfile'])
            if not sbcheck:
                raise FileNotFoundError(("WARNING: Superbias pipeline step configuration file not provided. "
                                        "This file is needed to run the pipeline."))
            refpixcheck = self.check_run_step(self.params['newRamp']['refpix_configfile'])
            if not refpixcheck:
                raise FileNotFoundError(("WARNING: Refpix pipeline step configuration file not provided. "
                                         "This file is needed to run the pipeline."))
            lincheck = self.check_run_step(self.params['newRamp']['linear_configfile'])
            if not lincheck:
                raise FileNotFoundError(("WARNING: Linearity pipeline step configuration file not provided. "
                                         "This file is needed to run the pipeline."))

    def check_run_step(self, filename):
        """Check to see if a filename exists in the parameter file
        or if it is set to none.

        Parameters
        ----------
        filename : str
            Name of the entry in the yaml file

        Returns
        -------
        state : bool
            True if entry is a filename, False if it is non or empty
        """
        if ((len(filename) == 0) or (filename.lower() == 'none')):
            return False
        else:
            return True

    def crop_dark(self, model):
        """Cut the dark current array down to the size dictated
        by the subarray bounds

        Parameters
        ----------
        model : obj
            read_fits object holding dark current as well as superbias and refpix signal

        Returns
        -------
        model : obj
            read_fits object with cropped data arrays
        """
        modshape = model.data.shape
        yd = modshape[-2]
        xd = modshape[-1]

        if ((self.subarray_bounds[0] != 0) or (self.subarray_bounds[2] != (xd - 1))
           or (self.subarray_bounds[1] != 0) or (self.subarray_bounds[3] != (yd - 1))):

            if len(modshape) == 4:
                model.data = model.data[:, :, self.subarray_bounds[1]:self.subarray_bounds[3] + 1,
                                        self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                try:
                    model.sbAndRefpix = model.sbAndRefpix[:, :, self.subarray_bounds[1]:
                                                          self.subarray_bounds[3] + 1,
                                                          self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                except:
                    pass
                try:
                    model.zeroframe = model.zeroframe[:, self.subarray_bounds[1]:
                                                      self.subarray_bounds[3] + 1,
                                                      self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                except:
                    pass
            if len(modshape) == 3:
                model.data = model.data[:, self.subarray_bounds[1]:self.subarray_bounds[3] + 1,
                                        self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                try:
                    model.sbAndRefpix = model.sbAndRefpix[:, self.subarray_bounds[1]:
                                                          self.subarray_bounds[3] + 1,
                                                          self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                except:
                    pass
                try:
                    model.zeroframe = model.zeroframe[:, self.subarray_bounds[1]:
                                                      self.subarray_bounds[3] + 1,
                                                      self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                except:
                    pass
            elif len(modshape) == 2:
                model.data = model.data[self.subarray_bounds[1]:self.subarray_bounds[3] + 1,
                                        self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                try:
                    model.sbAndRefpix = model.sbAndRefpix[self.subarray_bounds[1]:self.subarray_bounds[3] + 1,
                                                          self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                except:
                    pass
                try:
                    model.zeroframe = model.zeroframe[self.subarray_bounds[1]:
                                                      self.subarray_bounds[3] + 1,
                                                      self.subarray_bounds[0]:self.subarray_bounds[2] + 1]
                except:
                    pass

        # Make sure that if the output is supposedly a
        # 4-amplifier file, that the number of
        # pixels in the x direction is a multiple of 4.
        nfast = self.subarray_bounds[2] - self.subarray_bounds[0] + 1
        nslow = self.subarray_bounds[3] - self.subarray_bounds[1] + 1
        nramp = (self.params['Readout']['nframe'] + self.params['Readout']['nskip']) * \
            self.params['Readout']['ngroup']

        if self.params['Readout']['namp'] != 4 and self.params['Readout']['namp'] != 1:
            raise ValueError(('ERROR: amplifier mode specified ({}) is not allowed'
                              .format(self.params['Readout']['namp'])))

        if self.params['Readout']['namp'] == 4:
            n = int(nfast / 4) * 4
            if n != nfast:
                raise ValueError(("ERROR: 4 amplifier mode specified but the number of pixels "
                                  "in the fast read direction ({}) is not a multiple of 4."
                                  .format(nfast)))
        return model

    def darkints(self):
        """Check the number of integrations in the dark
        current file and compare with the requested
        number of integrations. Add/remove integrations
        if necessary

        Parameters
        ----------

        None


        Returns
        -------

        None
        """
        ndarkints, ngroups, ny, nx = self.dark.data.shape
        reqints = self.params['Readout']['nint']

        if reqints <= ndarkints:
            # Fewer requested integrations than are in the
            # input dark.
            print("{} output integrations requested.".format(reqints))
            self.dark.data = self.dark.data[0:reqints, :, :, :]
            if self.dark.sbAndRefpix is not None:
                self.dark.sbAndRefpix = self.dark.sbAndRefpix[0:reqints, :, :, :]
            if self.dark.zeroframe is not None:
                self.dark.zeroframe = self.dark.zeroframe[0:reqints, :, :]

        elif reqints > ndarkints:
            # More requested integrations than are in input dark.
            print(('Requested output has {} integrations, while input dark has only {}.'
                   .format(reqints, ndarkints)))
            print('Adding integrations to input dark by making copies of the input.')
            self.integration_copy(reqints, ndarkints)

    def data_volume_check(self, obj):
        """Make sure that the input integration has
        enough frames/groups to create the requested
        number of frames/groups of the output

        Parameters
        ----------
        obj : obj
            Instance of read_fits class containing dark current data and other info

        Returns
        -------
        None
        """
        ngroup = int(self.params['Readout']['ngroup'])
        nframe = int(self.params['Readout']['nframe'])
        nskip = int(self.params['Readout']['nskip'])

        inputframes = obj.data.shape[1]
        if ngroup * (nskip + nframe) > inputframes:
            print(("WARNING: Not enough frames in the input integration to "
                   "create the requested number of output groups. Input has "
                   "{} frames. Requested output is {} groups each created from "
                   "{} frames plus skipping {} frames between groups for a total "
                   "of {} frames."
                   .format(inputframes, ngroup, nframe, nskip, ngroup * (nframe + nskip))))
            print(("Making copies of {} dark current frames and adding them to "
                   "the end of the dark current integration."
                   .format(ngroup * (nskip + nframe) - inputframes)))

            # Figure out how many more frames we need,
            # in terms of how many copies of the original dark
            div = floor((ngroup * (nskip + nframe)) / inputframes)
            mod = (ngroup * (nskip + nframe)) % inputframes

            # If more frames are needed than there are frames
            # in the original dark, then make copies of the
            # entire thing as many times as necessary, adding
            # the signal from the previous final frame to each.
            for ints in range(div - 1):
                extra_frames = np.copy(obj.data)
                obj.data = np.hstack((obj.data, extra_frames + obj.data[:, -1, :, :]))
                if obj.sbAndRefpix is not None:
                    extra_sb = np.copy(obj.sbAndRefpix)
                    obj.sbAndRefpix = np.hstack((obj.sbAndRefpix, extra_sb + obj.sbAndRefpix[:, -1, :, :]))

            # At this point, if more frames are needed, but fewer
            # than an entire copy of self.dark.data, then add the
            # appropriate number of frames here.
            extra_frames = np.copy(obj.data[:, 1:mod + 1, :, :]) - obj.data[:, 0, :, :]
            obj.data = np.hstack((obj.data, extra_frames + obj.data[:, -1, :, :]))
            if obj.sbAndRefpix is not None:
                extra_sb = np.copy(obj.sbAndRefpix[:, 1:mod + 1, :, :]) - obj.sbAndRefpix[:, 0, :, :]
                obj.sbAndRefpix = np.hstack((obj.sbAndRefpix, extra_sb + obj.sbAndRefpix[:, -1, :, :]))

        elif ngroup * (nskip + nframe) < inputframes:
            # If there are more frames in the dark than we'll need,
            # crop the extras in order to reduce memory use
            obj.data = obj.data[:, 0:ngroup * (nskip + nframe), :, :]
            if obj.sbAndRefpix is not None:
                obj.sbAndRefpix = obj.sbAndRefpix[:, 0:ngroup * (nskip + nframe), :, :]
        obj.header['NGROUPS'] = ngroup * (nskip + nframe)

    def filecheck(self):
        """Make sure the requested input files exist
        or reference files, assume first that they are located in
        the directory tree under the datadir (from the MIRAGE_DATA
        environment variable). If not, assume the input is a full path
        and check there.
        """
        rlist = [['Reffiles', 'dark'],
                 ['Reffiles', 'linearized_darkfile'],
                 ['Reffiles', 'superbias'],
                 ['Reffiles', 'linearity'],
                 ['Reffiles', 'saturation']]
        for ref in rlist:
            self.ref_check(ref)

    def file_splits(self):
        """Check to see whether the final dark file will need to be
        split into segments due to large file size
        """
        # Let's declare the equivalent of a 100 group full frame ramp
        # as the upper limit for observation file size
        pixel_limit = 2048 * 2048 * 100

        xd = self.subarray_bounds[3] - self.subarray_bounds[1]
        yd = self.subarray_bounds[2] - self.subarray_bounds[0]
        pix_per_int = self.numgroups * yd * xd
        observation = pix_per_int * self.numints

        split = False
        int_list = [0, self.numints]
        if observation > pixel_limit:
            split = True
            ints_per_split = np.int(pixel_limit / pix_per_int)
            #num_splits = self.numints / ints_per_split
            int_list = np.arange(0, self.numints, ints_per_split)
            if int_list[-1] != self.numints:
                int_list = np.append(int_list, self.numints)
        return split, int_list

    def get_base_dark(self, input_file):
        """Read in the dark current ramp that will serve as the
        base for the simulated ramp"""

        # First make sure that the file exists
        local = os.path.isfile(input_file)
        # Future improvement: download a missing dark file
        if not local:
            try:
                print("Local copy of dark current file")
                print("{} not found.".format(input_file))
                print("Downloading.")
                darkfile = os.path.split(input_file)[1]
                hh = fits.open(os.path.join(self.dark_url, darkfile))
                hh.writeto(input_file)
            except Exception:
                print("Local copy of dark current file")
                print("{} not found.".format(input_file))
                print("And unable to download. Quitting.")

        self.dark = read_fits.Read_fits()
        self.dark.file = input_file

        # Depending on the method indicated in the input file
        # read in the dark using astropy or RampModel
        if self.params['Inst']['use_JWST_pipeline']:
            self.dark.read_datamodel()

            try:
                # Remove non-pipeline related keywords
                # (e.g. CV3 temps/voltages)
                self.dark.__delattr__('extra_fits')
            except:
                pass

        else:
            self.dark.read_astropy()

        # We assume that the input dark current integration is raw, which means
        # the data are in the original ADU measured by the detector. So the
        #  data range should be no larger than 65536
        darkrange = 1.*self.dark.data.max() - 1.*self.dark.data.min()
        if darkrange > 65535.:
            raise ValueError(("WARNING: Range of data values in the input dark is too large. "
                              "We assume the input dark is raw ADU values, with a range no more than 65536."))

        # If the inputs are signed integers, change to unsigned.
        if self.dark.data.min() < 0.:
            self.dark.data += 32768

        # If the input is any readout pattern other than RAPID, then
        # make sure that the output readout patten matches. Only RAPID
        # can be averaged and transformed into another readout pattern
        if self.dark.header['READPATT'] not in ['RAPID', 'NISRAPID', 'FGSRAPID']:
            if self.params['Readout']['readpatt'].upper() != self.dark.header['READPATT']:
                print(("WARNING: cannot transform input {} integration into "
                       "output {} integration.".format(self.dark.header['READPATT'],
                                                       self.params['Readout']['readpatt'])))
                raise ValueError(("Only RAPID, NISRAPID, or FGSRAPID inputs can be translated to a "
                                  "different readout pattern"))
        else:
            pass

        # Finally, collect information about the detector, which will be needed for astrometry later
        self.detector = self.dark.header['DETECTOR']
        self.instrument = self.dark.header['INSTRUME']
        self.fastaxis = self.dark.header['FASTAXIS']
        self.slowaxis = self.dark.header['SLOWAXIS']

    def get_frame_count_info(self):
        """Calculate information on the number of frames per group and
        per integration
        """
        self.numints = self.params['Readout']['nint']
        self.numgroups = self.params['Readout']['ngroup']
        numframes = self.params['Readout']['nframe']
        numskips = self.params['Readout']['nskip']
        numresets = self.params['Readout']['resets_bet_ints']

        self.frames_per_group = numframes + numskips
        self.frames_per_integration = self.numgroups * self.frames_per_group
        self.total_frames = self.numgroups * self.frames_per_group

        if self.numints > 1:
            # Frames for all integrations
            self.total_frames *= self.numints
            # Don't worry about counting reset frames between integrations
            # We're not concerned with timing here.

    def integration_copy(self, req, darkint):
        """Use copies of integrations in the dark current input to
        make up integrations in the case where the output has
        more integrations than the input

        Parameters
        ----------
        req : int
            Requested number of dark current integrations for the output

        darkint : int
            Number of integrations in the input dark current exposure

        Returns
        -------
        None
        """

        ncopies = np.int((req - darkint) / darkint)
        extras = (req - darkint) % darkint

        # Full copies of the dark exposure (all integrations)
        if ncopies > 0:
            copydata = self.dark.data
            if self.dark.sbAndRefpix is not None:
                copysb = self.dark.sbAndRefpix
            if self.dark.zeroframe is not None:
                copyzero = self.dark.zeroframe
            for i in range(ncopies):
                self.dark.data = np.vstack((self.dark.data, copydata))
                if self.dark.sbAndRefpix is not None:
                    self.dark.sbAndRefpix = np.vstack((self.dark.sbAndRefpix, copysb))
                if self.dark.zeroframe is not None:
                    self.dark.zeroframe = np.vstack((self.dark.zeroframe, copyzero))

        # Partial copy of dark (some integrations)
        if extras > 0:
            self.dark.data = np.vstack((self.dark.data, self.dark.data[0:extras, :, :, :]))
            if self.dark.sbAndRefpix is not None:
                self.dark.sbAndRefpix = np.vstack((self.dark.sbAndRefpix,
                                                   self.dark.sbAndRefpix[0:extras, :, :, :]))
            if self.dark.zeroframe is not None:
                self.dark.zeroframe = np.vstack((self.dark.zeroframe, self.dark.zeroframe[0:extras, :, :]))

        self.dark.header['NINTS'] = req

    def linearize_dark(self, darkobj):
        """Beginning with the input dark current ramp, run the dq_init, saturation, superbias
        subtraction, refpix and nonlin pipeline steps in order to produce a linearized
        version of the ramp. This will be used when combining the dark ramp with the
        simulated signal ramp.

        Parameters
        -----------
        darkobj : obj
            Instance of read_fits class containing dark current data and info

        Returns
        -------
        linDarkObj : obj
            Modified read_fits instance with linearized dark current data
        """
        from jwst.dq_init import DQInitStep
        from jwst.saturation import SaturationStep
        from jwst.superbias import SuperBiasStep
        from jwst.refpix import RefPixStep
        from jwst.linearity import LinearityStep

        # First we need to place the read_fits object into a RampModel instance
        if self.runStep['linearized_darkfile']:
            subfile = self.params['Reffiles']['linearized_darkfile']
        else:
            subfile = self.params['Reffiles']['dark']
        dark = darkobj.insert_into_datamodel(subfile)

        print('Creating a linearized version of the dark current input ramp')
        print('using JWST calibration pipeline.')

        # Run the DQ_Init step
        if self.runStep['badpixmask']:
            linDark = DQInitStep.call(dark,
                                      config_file=self.params['newRamp']['dq_configfile'],
                                      override_mask=self.params['Reffiles']['badpixmask'])
        else:
            linDark = DQInitStep.call(dark, config_file=self.params['newRamp']['dq_configfile'])

        # If the saturation map is provided, use it. If not, default to whatever is in CRDS
        if self.runStep['saturation_lin_limit']:
            linDark = SaturationStep.call(linDark,
                                          config_file=self.params['newRamp']['sat_configfile'],
                                          override_saturation=self.params['Reffiles']['saturation'])
        else:
            linDark = SaturationStep.call(linDark,
                                          config_file=self.params['newRamp']['sat_configfile'])

        # If the superbias file is provided, use it. If not, default to whatever is in CRDS
        if self.runStep['superbias']:
            linDark = SuperBiasStep.call(linDark,
                                         config_file=self.params['newRamp']['superbias_configfile'],
                                         override_superbias=self.params['Reffiles']['superbias'])
        else:
            linDark = SuperBiasStep.call(linDark,
                                         config_file=self.params['newRamp']['superbias_configfile'])

        # Reference pixel correction
        linDark = RefPixStep.call(linDark,
                                  config_file=self.params['newRamp']['refpix_configfile'])

        # Save a copy of the superbias- and reference pixel-subtracted
        # dark. This will be used later to add these effects back in
        # after the synthetic signals have been added and the non-linearity
        # effects are added back in when using the PROPER combine method.
        sbAndRefpixEffects = dark.data - linDark.data

        # Linearity correction - save the output so that you won't need to
        # re-run the pipeline when using the same dark current file in the
        # future. Use the linearity coefficient file if provided
        base_name = self.params['Output']['file'].split('/')[-1]
        linearoutfile = base_name[0:-5] + '_linearized_dark_current_ramp.fits'
        if self.runStep['linearity']:
            linDark = LinearityStep.call(linDark,
                                         config_file=self.params['newRamp']['linear_configfile'],
                                         override_linearity=self.params['Reffiles']['linearity'],
                                         output_file=linearoutfile, save_results=True,
                                         output_dir=self.params['Output']['directory'])
        else:
            linDark = LinearityStep.call(linDark,
                                         config_file=self.params['newRamp']['linear_configfile'],
                                         output_file=linearoutfile, save_results=True,
                                         output_dir=self.params['Output']['directory'])

        print(("Linearized dark (output directly from pipeline saved as {}"
               .format(os.path.join(self.params['Output']['directory'], linearoutfile))))

        # Now we need to put the data back into a read_fits object
        linDarkobj = read_fits.Read_fits()
        linDarkobj.model = linDark
        linDarkobj.rampmodel_to_obj()
        linDarkobj.sbAndRefpix = sbAndRefpixEffects

        return linDarkobj

    def prepare(self):
        """MAIN FUNCTION"""
        print("\nRunning dark_prep....\n")

        # Read in the yaml parameter file
        self.read_parameter_file()

        # Create dictionary to use when looking in CRDS for reference files
        self.crds_dict = crds_tools.dict_from_yaml(self.params)

        # Expand param entries to full paths where appropriate
        self.params = utils.full_paths(self.params, self.modpath, self.crds_dict, offline=self.offline)
        self.filecheck()

        # Base name for output files
        base_name = self.params['Output']['file'].split('/')[-1]
        self.basename = os.path.join(self.params['Output']['directory'],
                                     base_name[0:-5])

        # Check the entered read pattern info
        self.readpattern_check()

        # Check input parameters for any bad values
        self.check_params()

        # Find out how many groups/integrations we need
        self.get_frame_count_info()

        # If the simulation will have more than one integration, generate
        # a list of dark current files that can be read in to provide
        # more data
        if self.numints > 1:
            if self.runStep['linearized_darkfile']:
                dark_dir = os.path.split(self.params['Reffiles']['linearized_darkfile'])[0]
            else:
                dark_dir = os.path.split(self.params['Reffiles']['dark'])[0]
            dark_list = glob(os.path.join(dark_dir, '*.fits'))
        else:
            if self.runStep['linearized_darkfile']:
                dark_list = [self.params['Reffiles']['linearized_darkfile']]
            else:
                dark_list = [self.params['Reffiles']['dark']]
        dark_list = np.array(dark_list)

        # Read in the subarray definition file
        self.subdict = utils.read_subarray_definition_file(self.params['Reffiles']['subarray_defs'])
        self.params = utils.get_subarray_info(self.params, self.subdict)

        # Get the subarray boundaries from pysiaf
        siaf_inst = self.params['Inst']['instrument']
        instrument_siaf = siaf_interface.get_instance(siaf_inst)
        self.siaf = instrument_siaf[self.params['Readout']['array_name']]
        junk0, junk1, self.ffsize, \
            self.subarray_bounds = siaf_interface.get_siaf_information(instrument_siaf,
                                                                       self.params['Readout']['array_name'],
                                                                       0.0, 0.0,
                                                                       self.params['Telescope']['rotation'])

        # If there will be too many frames, then the file will need
        # to be split into pieces to save memory
        xdim = self.subarray_bounds[2] - self.subarray_bounds[0] + 1
        ydim = self.subarray_bounds[3] - self.subarray_bounds[1] + 1
        split_seed, group_segment_indexes, integration_segment_indexes = find_file_splits(xdim, ydim,
                                                                                          self.params['Readout']['ngroup'],
                                                                                          self.params['Readout']['nint'])
        # In order to avoid the case of having a single integration
        # in the final file, which leads to rate rather than rateints
        # files in the pipeline, check to be sure that the integration
        # splitting indexes indicate the last split isn't a single
        # integration
        if len(integration_segment_indexes) > 2:
            delta_int = integration_segment_indexes[1:] - integration_segment_indexes[0: -1]
            if delta_int[-1] == 1 and delta_int[0] != 1:
                integration_segment_indexes[-2] -= 1
                print('Adjusted to avoid single integration: ', integration_segment_indexes)

        # This currently pays attention only to splitting the file on integrations, and not
        # within an integration. It seems like we can keep it this way as the dark data is
        # reordered to the requested readout pattern before saving. It seems unlikely
        # that observations will be taken that require splitting the file within integrations
        # in that case. BUT at one point we do still have all the frames in memory
        # (within data_volume_check). So at that point, a DEEP exposure with 20 groups
        # will have 400 frames open. This is also an unlikely situation, but it could
        # be a problem.
        self.dark_files = []
        for seg_index, segment in enumerate(integration_segment_indexes[:-1]):
            # Get the number of integrations being simulated
            if split_seed:
                number_of_ints = integration_segment_indexes[seg_index+1] - segment
            else:
                number_of_ints = self.numints

            print('Segment number: ', seg_index)
            print('Number of integrations:', number_of_ints)

            # Generate the list of dark current files to use
            if number_of_ints == 1:
                use_all_files = False
                files_to_use = [dark_list[0]]
            elif number_of_ints > 1 and number_of_ints <= len(dark_list):
                use_all_files = False
                try:
                    random_indices = np.random.choice(len(dark_list), number_of_ints, replace=False)
                    files_to_use = dark_list[random_indices]
                except TypeError:
                    print(dark_list)
                    print(number_of_ints)
                    print(random_indices)
                    print(type(random_indices))
                    print(type(random_indices[0]))
                    raise ValueError("Error in random choice.")
            else:
                use_all_files = True
                files_to_use = dark_list

            print('Dark files to use:')
            print(files_to_use)

            # Create a list describing which dark current file goes with
            # which integration. Force the 0th element to be the first
            # file in the list, to be assured that the final_dark variable
            # is defined.
            mapping = np.random.choice(len(files_to_use), size=number_of_ints)
            mapping[0] = 0

            # Loop over dark current files
            for file_index, filename in enumerate(files_to_use):

                print('Working on dark file: {}'.format(filename))
                if not use_all_files:
                    frames = np.arange(len(files_to_use))
                else:
                    # In this case, we need to repeat at least some of
                    # the dark ramps because the number of integrations
                    # is larger than the number of dark current files
                    frames = np.where(mapping == file_index)[0]

                # If the random number picker didn't pick this file for
                # any integrations, then skip reading it in.
                if len(frames) == 0:
                    continue

                # Read in the input dark current frame
                if not self.runStep['linearized_darkfile']:
                    self.get_base_dark(filename)
                    self.linDark = None
                else:
                    self.read_linear_dark(filename)
                    self.dark = self.linDark

                # Make sure there is enough data (frames/groups)
                # in the input integration to produce
                # the proposed output integration
                self.data_volume_check(self.dark)

                # Put the input dark (or linearized dark) into the
                # requested readout pattern
                self.dark, sbzeroframe = self.reorder_dark(self.dark)
                print(('DARK has been reordered to {} to match the input readpattern of {}'
                       .format(self.dark.data.shape, self.dark.header['READPATT'])))

                # If a raw dark was read in, create linearized version
                # here using the SSB pipeline. Better to do this
                # on the full dark before cropping, so that reference
                # pixels can be used in the processing.
                if ((self.params['Inst']['use_JWST_pipeline']) & (self.runStep['linearized_darkfile'] is False)):

                    # Linearize the dark ramp via the SSB pipeline.
                    # Also save a diff image of the original dark minus
                    # the superbias and refpix subtracted dark, to use later.

                    # In order to linearize the dark, the JWST pipeline must
                    # be present, and self.dark will have to be translated back
                    # into a RampModel instance
                    # print('Working on {}'.format(self.dark))
                    self.linDark = self.linearize_dark(self.dark)
                    print("Linearized dark shape: {}".format(self.linDark.data.shape))

                    if self.params['Readout']['readpatt'].upper() in ['RAPID', 'NISRAPID', 'FGSRAPID']:
                        print(("Output is {}, grabbing zero frame from linearized dark"
                               .format(self.params['Readout']['readpatt'].upper())))
                        self.zeroModel = read_fits.Read_fits()
                        self.zeroModel.data = self.linDark.data[:, 0, :, :]
                        self.zeroModel.sbAndRefpix = self.linDark.sbAndRefpix[:, 0, :, :]
                    elif ((self.params['Readout']['readpatt'].upper() not in ['RAPID', 'NISRAPID', 'FGSRAPID']) &
                          (self.dark.zeroframe is not None)):
                        print("Now we need to linearize the zeroframe because the")
                        print("output readpattern is not RAPID, NISRAPID, or FGSRAPID")
                        # Now we need to linearize the zeroframe. Place it
                        # into a RampModel instance before running the
                        # pipeline steps
                        self.zeroModel = read_fits.Read_fits()
                        self.zeroModel.data = np.expand_dims(self.dark.zeroframe, axis=1)
                        self.zeroModel.header = self.linDark.header
                        self.zeroModel.header['NGROUPS'] = 1
                        self.zeroModel = self.linearize_dark(self.zeroModel)
                        # Return the zeroModel data to 3 dimensions
                        # integrations, y, x
                        self.zeroModel.data = self.zeroModel.data[:, 0, :, :]
                        self.zeroModel.sbAndRefpix = self.zeroModel.sbAndRefpix[:, 0, :, :]
                        # In this case the zeroframe has changed from what
                        # was read in. So let's remove the original zeroframe
                        # to avoid confusion
                        self.linDark.zeroframe = np.zeros(self.linDark.zeroframe.shape)
                    else:
                        self.zeroModel = None

                    # Now crop self.linDark, self.dark, and zeroModel
                    # to requested subarray
                    self.dark = self.crop_dark(self.dark)
                    self.linDark = self.crop_dark(self.linDark)

                    if self.zeroModel is not None:
                        self.zeroModel = self.crop_dark(self.zeroModel)

                elif self.runStep['linearized_darkfile']:
                    # If no pipeline is run
                    self.zeroModel = read_fits.Read_fits()
                    self.zeroModel.data = self.dark.zeroframe
                    self.zeroModel.sbAndRefpix = sbzeroframe

                    # Crop the linearized dark to the requested
                    # subarray size
                    # THIS WILL CROP self.dark AS WELL SINCE
                    # self.linDark IS JUST A REFERENCE IN THE NON
                    # PIPELINE CASE!!
                    self.linDark = self.crop_dark(self.linDark)
                    if self.zeroModel.data is not None:
                        self.zeroModel = self.crop_dark(self.zeroModel)
                else:
                    raise NotImplementedError(("Mode not yet supported! Must use either: use_JWST_pipeline "
                                               "= True and a raw or linearized dark or supply a linearized dark. "
                                               "Cannot yet skip the pipeline and provide a raw dark."))

                if file_index == 0:
                    #print('self.linDark shape is {}. Expecting it to be 4D'.format(self.linDark.data.shape))
                    junk, num_grps, ydim, xdim = self.linDark.data.shape
                    final_dark = np.zeros((number_of_ints, num_grps, ydim, xdim))
                    final_sbandrefpix = np.zeros((number_of_ints, num_grps, ydim, xdim))
                    final_zerodata = np.zeros((number_of_ints, ydim, xdim))
                    final_zero_sbandrefpix = np.zeros((number_of_ints, ydim, xdim))

                if not use_all_files:
                    print('Setting integration {} to use file {}\n'.format(file_index, os.path.split(filename)[1]))
                    final_dark[file_index, :, :, :] = self.linDark.data
                    final_sbandrefpix[file_index, :, :, :] = self.linDark.sbAndRefpix
                    final_zerodata[file_index, :, :] = self.zeroModel.data
                    final_zero_sbandrefpix[file_index, :, :] = self.zeroModel.sbAndRefpix
                else:
                    # In this case, we need to repeat at least some of
                    # the dark ramps because the number of integrations
                    # is larger than the number of dark current files
                    #frames = np.where(mapping == file_index)[0]
                    print('File number {} will be used for integrations {}'.format(file_index, frames))
                    final_dark[frames, :, :, :] = self.linDark.data

                    #final_sbandrefpix[frames, :, :, :] = self.linDark.sbAndRefpix
                    for frame in frames:
                        final_sbandrefpix[frame, :, :, :] = self.linDark.sbAndRefpix


                    final_zerodata[frames, :, :] = self.zeroModel.data
                    final_zero_sbandrefpix[frames, :, :] = self.zeroModel.sbAndRefpix

            # Save the linearized dark
            # if self.params['Output']['save_intermediates']:
            h0 = fits.PrimaryHDU()
            h1 = fits.ImageHDU(final_dark, name='SCI')
            h2 = fits.ImageHDU(final_sbandrefpix, name='SBANDREFPIX')
            h3 = fits.ImageHDU(final_zerodata, name='ZEROFRAME')
            h4 = fits.ImageHDU(final_zero_sbandrefpix, name='ZEROSBANDREFPIX')

            # Populate basic info in the 0th extension header
            nints, ngroups, yd, xd = final_dark.shape
            h0.header['READPATT'] = self.params['Readout']['readpatt'].upper()
            h0.header['NINTS'] = nints
            h0.header['NGROUPS'] = ngroups
            h0.header['NFRAMES'] = self.params['Readout']['nframe']
            h0.header['NSKIP'] = self.params['Readout']['nskip']
            h0.header['DETECTOR'] = self.detector
            h0.header['INSTRUME'] = self.instrument
            h0.header['SLOWAXIS'] = self.slowaxis
            h0.header['FASTAXIS'] = self.fastaxis

            # Add some basic Mirage-centric info
            h0.header['MRGEVRSN'] = (mirage.__version__, 'Mirage version used')
            h0.header['YAMLFILE'] = (self.paramfile, 'Mirage input yaml file')

            hl = fits.HDUList([h0, h1, h2, h3, h4])
            if split_seed:
                objname = self.basename + '_seg{}_linear_dark_prep_object.fits'.format(str(seg_index+1).zfill(3))
            else:
                objname = self.basename + '_linear_dark_prep_object.fits'
            objname = os.path.join(self.params['Output']['directory'], objname)
            hl.writeto(objname, overwrite=True)
            print(("Linearized dark frame plus superbias and reference"
                   "pixel signals, as well as zeroframe, saved to {}. "
                   "This can be used as input to the observation "
                   "generator.".format(objname)))
            self.dark_files.append(objname)

        # If only one dark current file is needed, return just the file
        # name rather than a list.
        if len(self.dark_files) == 1:
            self.dark_files = self.dark_files[0]

        # important variables
        # self.linDark
        # self.zeroModel
        # Make a read_fits instance that contains all the important
        # information, for ease in connecting with next step of the
        # simulator
        self.prepDark = read_fits.Read_fits()
        self.prepDark.data = final_dark
        self.prepDark.zeroframe = final_zerodata
        self.prepDark.sbAndRefpix = final_sbandrefpix
        self.prepDark.zero_sbAndRefpix = final_zero_sbandrefpix
        self.prepDark.header = self.linDark.header

    def read_linear_dark(self, input_file):
        """Read in the linearized version of the dark current ramp
        using the read_fits class"""
        try:
            print(('Reading in linearized dark current ramp from {}'
                   .format(input_file)))
            self.linDark = read_fits.Read_fits()
            #self.linDark.file = self.params['Reffiles']['linearized_darkfile']
            self.linDark.file = input_file
            self.linDark.read_astropy()
        except:
            raise IOError('WARNING: Unable to read in linearized dark ramp.')

        # Finally, collect information about the detector, which will be needed for astrometry later
        self.detector = self.linDark.header['DETECTOR']
        self.instrument = self.linDark.header['INSTRUME']
        self.fastaxis = self.linDark.header['FASTAXIS']
        self.slowaxis = self.linDark.header['SLOWAXIS']

    def read_parameter_file(self):
        """Read in the yaml parameter file"""
        try:
            with open(self.paramfile, 'r') as infile:
                self.params = yaml.safe_load(infile)
        except FileNotFoundError:
            print("WARNING: unable to open {}".format(self.paramfile))

    def readpattern_check(self):
        """Check the readout pattern that's entered and set the number of averaged frames
        as well as the number of skipped frames per group"""
        self.params['Readout']['readpatt'] = self.params['Readout']['readpatt'].upper()

        # Read in readout pattern definition file
        # and make sure the possible readout patterns are in upper case
        self.readpatterns = ascii.read(self.params['Reffiles']['readpattdefs'])
        self.readpatterns['name'] = [s.upper() for s in self.readpatterns['name']]

        # If the requested readout pattern is in the table of options,
        # then adopt the appropriate nframe and nskip
        if self.params['Readout']['readpatt'] in self.readpatterns['name']:
            mtch = self.params['Readout']['readpatt'] == self.readpatterns['name']
            self.params['Readout']['nframe'] = self.readpatterns['nframe'][mtch].data[0]
            self.params['Readout']['nskip'] = self.readpatterns['nskip'][mtch].data[0]
            print(('Requested readout pattern {} is valid. '
                  'Using the nframe = {} and nskip = {}'
                   .format(self.params['Readout']['readpatt'],
                           self.params['Readout']['nframe'],
                           self.params['Readout']['nskip'])))
        else:
            # If the read pattern is not present in the definition file
            # then quit.
            raise ValueError(("WARNING: the {} readout pattern is not defined in {}."
                              .format(self.params['Readout']['readpatt'],
                                      self.params['Reffiles']['readpattdefs'])))

    def ref_check(self, rele):
        """Check for the existence of the input reference file
        Assume first that the file is in the directory tree
        specified by the MIRAGE_DATA environment variable.

        Parameters
        ----------
        rele : tup
            Keywords needed to access a given entry in self.params from the
            input yaml file

        Returns
        -------
        None
        """
        rfile = self.params[rele[0]][rele[1]]
        if rfile.lower() != 'none':
            c1 = os.path.isfile(rfile)
            if not c1:
                raise FileNotFoundError(("WARNING: Unable to locate the {}, {} "
                                         "input file! Not present in {}"
                                         .format(rele[0], rele[1], rfile)))

    def reorder_dark(self, dark):
        """Reorder the input dark ramp using the requested
        readout pattern (nframe, nskip). If the initial
        dark ramp is RAPID, NISRAPID, or FGSRAPID, then save and return
        the 0th frame.

        Parameters
        ----------
        dark : obj
            Instance of read_fits class containing dark current data and info

        Returns
        -------
        dark : obj
            Instance of read_fits class where dark data have been reordered to match
            the requested readout pattern

        sbzero : numpy.ndarray
            Zeroth frame data from the dark current data. This is saved separately
            because averaging for non-RAPID readout patterns will destroy the frame
        """
        if self.params['Reffiles']['linearized_darkfile']:
            datatype = np.float
        else:
            datatype = np.int32

        # Get the info for the dark integration
        darkpatt = dark.header['READPATT']
        dark_nframe = dark.header['NFRAMES']
        mtch = self.readpatterns['name'].data == darkpatt
        dark_nskip = self.readpatterns['nskip'].data[mtch][0]

        nint, ngroup, yd, xd = dark.data.shape
        #outdark = np.zeros((self.params['Readout']['nint'], self.params['Readout']['ngroup'], yd, xd))
        outdark = np.zeros((1, self.params['Readout']['ngroup'], yd, xd))

        if dark.sbAndRefpix is not None:
            #outsb = np.zeros((self.params['Readout']['nint'], self.params['Readout']['ngroup'], yd, xd))
            outsb = np.zeros((1, self.params['Readout']['ngroup'], yd, xd))

        # We can only keep a zero frame around if the input dark
        # is RAPID, NISRAPID, or FGSRAPID. Otherwise that information is lost.
        darkzero = None
        sbzero = None
        rapids = ['RAPID', 'NISRAPID', 'FGSRAPID']
        if ((darkpatt in rapids) & (dark.zeroframe is None)):
            dark.zeroframe = dark.data[:, 0, :, :]
            print("Saving 0th frame from data to the zeroframe extension")
            if dark.sbAndRefpix is not None:
                sbzero = dark.sbAndRefpix[:, 0, :, :]
        elif ((darkpatt not in rapids) & (dark.zeroframe is None)):
            print("Unable to save the zeroth frame because the input dark current ramp is not RAPID.")
            sbzero = None
        elif ((darkpatt in rapids) & (dark.zeroframe is not None)):
            # In this case we already have the zeroframe
            if dark.sbAndRefpix is not None:
                sbzero = dark.sbAndRefpix[:, 0, :, :]
        elif ((darkpatt not in rapids) & (dark.zeroframe is not None)):
            # Non RAPID dark, zeroframe is present, but
            # we can't get sbAndRefpix for the zeroth frame
            # because the pattern is not RAPID.
            sbzero = None

        # We have already guaranteed that either the readpatterns match
        # or the dark is RAPID, so no need to worry about checking for
        # other cases here.

        if ((darkpatt in rapids) and (self.params['Readout']['readpatt'] not in rapids)):

            framesPerGroup = self.params['Readout']['nframe'] + self.params['Readout']['nskip']
            accumimage = np.zeros_like(outdark[0, 0, :, :], dtype=datatype)

            if dark.sbAndRefpix is not None:
                zeroaccumimage = np.zeros_like(outdark[0, 0, :, :], dtype=np.float)

            # Loop over integrations
            #for integ in range(self.params['Readout']['nint']):
            for integ in range(dark.data.shape[0]):
                frames = np.arange(0, self.params['Readout']['nframe'])

                # Loop over groups
                for i in range(self.params['Readout']['ngroup']):
                    # Average together the appropriate frames,
                    # skip the appropriate frames
                    print(("Averaging dark current ramp. Frames {}, to become "
                           "group {}".format(frames, i)))

                    # If averaging needs to be done
                    if self.params['Readout']['nframe'] > 1:
                        accumimage = np.mean(dark.data[integ, frames, :, :], axis=0)
                        if dark.sbAndRefpix is not None:
                            zeroaccumimage = np.mean(dark.sbAndRefpix[integ, frames, :, :], axis=0)

                        # If no averaging needs to be done
                    else:
                        accumimage = dark.data[integ, frames[0], :, :]
                        if dark.sbAndRefpix is not None:
                            zeroaccumimage = dark.sbAndRefpix[integ, frames[0], :, :]

                    outdark[integ, i, :, :] += accumimage
                    if dark.sbAndRefpix is not None:
                        outsb[integ, i, :, :] += zeroaccumimage

                    # Increment the frame indexes
                    frames = frames + framesPerGroup

        elif (self.params['Readout']['readpatt'] == darkpatt):
            # If the input dark is not RAPID, or if the readout
            # pattern of the input dark and
            # the output ramp match, then no averaging needs to
            # be done
            outdark = dark.data[:, 0:self.params['Readout']['ngroup'], :, :]
            if dark.sbAndRefpix is not None:
                outsb = dark.sbAndRefpix[:, 0:self.params['Readout']['ngroup'], :, :]
        else:
            # This check should already have been done,
            # but just to be sure...
            raise ValueError(("WARNING: dark current readout pattern is {} and requested "
                              "output is {}. Cannot convert between the two."
                              .format(darkpatt, self.params['Readout']['readpatt'])))

        # Now place the reorganized dark into the data object and
        # update the appropriate metadata
        dark.data = outdark
        if dark.sbAndRefpix is not None:
            dark.sbAndRefpix = outsb
        dark.header['READPATT'] = self.params['Readout']['readpatt']
        dark.header['NFRAMES'] = self.params['Readout']['nframe']
        dark.header['NSKIP'] = self.params['Readout']['nskip']
        dark.header['NGROUPS'] = self.params['Readout']['ngroup']
        dark.header['NINTS'] = self.params['Readout']['nint']
        return dark, sbzero

    def add_options(self, parser=None, usage=None):
        if parser is None:
            parser = argparse.ArgumentParser(usage=usage, description='Simulate JWST ramp')
        parser.add_argument("paramfile", help=("File describing the input parameters and instrument "
                                               "settings to use. (YAML format)."))
        return parser


if __name__ == '__main__':

    usagestring = 'USAGE: dark_prep.py inputs.yaml'

    indark = DarkPrep()
    parser = indark.add_options(usage=usagestring)
    args = parser.parse_args(namespace=indark)
    indark.prepare()



#! /usr/bin/env python
import glob
import io
import os
import yaml
import sys

from mirage import imaging_simulator
from mirage.yaml import yaml_generator


home_dir = os.environ['HOME']

ami_example_dir = os.path.dirname(__file__)
print(ami_example_dir)

xml_name = os.path.join(ami_example_dir, 'input_files/793_mirage_example.xml')
pointing_name = os.path.join(ami_example_dir, 'input_files/793_mirage_example.pointing')

output_directory = os.path.join(home_dir, 'ami_mirage_simulation_example')
simdata_output_directory = output_directory

# AB Dor is in field 19 and HD37093 is in field 20 in Kevin's targets.xlsx file
catalogues = {'AB-DOR': {'point_source': os.path.join(ami_example_dir,'stars_field19_20_combined_allfilters.list')
                        },
              'HD-37093': {'point_source': os.path.join(ami_example_dir,'stars_field19_20_combined_allfilters.list')
                        }
             }
pav3 = 275
dates = '2020-09-20'
reffile_defaults = 'crds'
datatype = 'linear, raw'


yam = yaml_generator.SimInput(input_xml=xml_name, pointing_file=pointing_name,
                              catalogs=catalogues, roll_angle=pav3,
                              dates=dates, reffile_defaults=reffile_defaults,
                              verbose=True, output_dir=output_directory,
                              simdata_output_dir=simdata_output_directory,
                              datatype=datatype)

yam.create_inputs()


#Create all files
yaml_files = glob.glob(os.path.join(output_directory, 'jw*.yaml'))
#create one file for testing
#yaml_files = glob.glob(os.path.join(output_directory, 'jw00793001001_01101_00001_nis.yaml'))
print(yaml_files)

for file in yaml_files:

    # set astrometric reference file to None to use pysiaf
    with open(file, 'r') as infile:
        yaml_content = yaml.load(infile)
    yaml_content['Reffiles']['astrometric'] = 'None'
    yaml_content['psf_wing_threshold_file'] = 'config'
    modified_file = file.replace('.yaml', '_mod.yaml')
    with io.open(modified_file, 'w') as outfile:
        yaml.dump(yaml_content, outfile, default_flow_style=False)

    t1 = imaging_simulator.ImgSim()
    t1.paramfile = str(modified_file)
    t1.create()




#! /usr/bin/env python

"""Download the reference files needed to run Mirage. Extract and unzip
files, and place into the proper directory. Inform the user how to set
the MIRAGE_DATA encironment variable.
"""
import os
import requests
import shutil
import tarfile
import gzip

from mirage.utils.utils import ensure_dir_exists


# Cosmic ray libraries
NIRCAM_CR_LIBRARY_URL = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/cosmic_ray_library/mirage_nircam_cr_library.tar.gz']
NIRISS_CR_LIBRARY_URL = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/cosmic_ray_library/mirage_niriss_cr_library.tar.gz']
FGS_CR_LIBRARY_URL = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/cosmic_ray_library/mirage_fgs_cr_library.tar.gz']

# Gridded PSF libraries
NIRCAM_GRIDDED_PSF_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/psf_libraries/gridded_psf_libraries/nircam_psf_wings_library.tar.gz',
                           'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/psf_libraries/gridded_psf_libraries/nircam_Amodule_det12_gridded_psf_library.tar.gz',
                           'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/psf_libraries/gridded_psf_libraries/nircam_Amodule_det345_gridded_psf_library.tar.gz',
                           'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/psf_libraries/gridded_psf_libraries/nircam_Bmodule_det12_gridded_psf_library.tar.gz',
                           'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/psf_libraries/gridded_psf_libraries/nircam_Bmodule_det345_gridded_psf_library.tar.gz',
                           ]
NIRISS_GRIDDED_PSF_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/psf_libraries/gridded_psf_libraries/niriss_gridded_psf_library.tar.gz']
FGS_GRIDDED_PSF_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/psf_libraries/gridded_psf_libraries/fgs_gridded_psf_library.tar.gz']

# Grism-related files
NIRCAM_GRISM_FILES = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/grism/nircam_grism_library.tar.gz']
NIRISS_GRISM_FILES = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/grism/niriss_grism_library.tar.gz']

# Dark current files
NIRCAM_RAW_DARK_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A1/NRCNRCA1-DARK-60082202011_1_481_SE_2016-01-09T00h03m58_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A1/NRCNRCA1-DARK-60090213141_1_481_SE_2016-01-09T02h53m12_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A1/NRCNRCA1-DARK-60090604481_1_481_SE_2016-01-09T06h52m47_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A1/NRCNRCA1-DARK-60091005411_1_481_SE_2016-01-09T10h56m36_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A1/NRCNRCA1-DARK-60091434481_1_481_SE_2016-01-09T15h50m45_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A2/NRCNRCA2-DARK-60082224241_1_482_SE_2016-01-09T00h10m36_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A2/NRCNRCA2-DARK-60090235001_1_482_SE_2016-01-09T04h17m03_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A2/NRCNRCA2-DARK-60090635511_1_482_SE_2016-01-09T07h05m19_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A2/NRCNRCA2-DARK-60091030561_1_482_SE_2016-01-09T11h03m17_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A2/NRCNRCA2-DARK-60091457131_1_482_SE_2016-01-09T15h50m45_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A3/NRCNRCA3-DARK-60082245481_1_483_SE_2016-01-09T00h04m26_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A3/NRCNRCA3-DARK-60090321241_1_483_SE_2016-01-09T04h17m10_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A3/NRCNRCA3-DARK-60090656591_1_483_SE_2016-01-09T07h31m27_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A3/NRCNRCA3-DARK-60091052561_1_483_SE_2016-01-09T11h28m06_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A3/NRCNRCA3-DARK-60091522581_1_483_SE_2016-01-09T16h30m34_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A4/NRCNRCA4-DARK-60082307391_1_484_SE_2016-01-09T00h04m08_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A4/NRCNRCA4-DARK-60090259591_1_484_SE_2016-01-09T04h16m50_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A4/NRCNRCA4-DARK-60090720041_1_484_SE_2016-01-09T07h58m26_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A4/NRCNRCA4-DARK-60091117441_1_484_SE_2016-01-09T11h52m23_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A4/NRCNRCA4-DARK-60091548131_1_484_SE_2016-01-09T16h30m50_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A5/NRCNRCALONG-DARK-60082329041_1_485_SE_2016-01-09T00h04m16_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A5/NRCNRCALONG-DARK-60090344021_1_485_SE_2016-01-09T04h16m42_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A5/NRCNRCALONG-DARK-60090746381_1_485_SE_2016-01-09T08h21m48_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A5/NRCNRCALONG-DARK-60091140151_1_485_SE_2016-01-09T14h23m49_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/A5/NRCNRCALONG-DARK-60091611271_1_485_SE_2016-01-09T17h16m35_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B1/NRCNRCB1-DARK-60082356471_1_486_SE_2016-01-09T02h47m00_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B1/NRCNRCB1-DARK-60090405201_1_486_SE_2016-01-09T05h33m56_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B1/NRCNRCB1-DARK-60090807581_1_486_SE_2016-01-09T08h48m11_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B1/NRCNRCB1-DARK-60091205311_1_486_SE_2016-01-09T14h30m08_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B1/NRCNRCB1-DARK-60091636021_1_486_SE_2016-01-09T17h16m13_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B2/NRCNRCB2-DARK-60090021181_1_487_SE_2016-01-09T02h51m54_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B2/NRCNRCB2-DARK-60090427541_1_487_SE_2016-01-09T05h33m14_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B2/NRCNRCB2-DARK-60090830131_1_487_SE_2016-01-09T08h59m50_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B2/NRCNRCB2-DARK-60091230011_1_487_SE_2016-01-09T14h23m47_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B2/NRCNRCB2-DARK-60091735131_1_487_SE_2016-01-09T18h09m45_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B3/NRCNRCB3-DARK-60090043151_1_488_SE_2016-01-09T02h53m21_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B3/NRCNRCB3-DARK-60090451471_1_488_SE_2016-01-09T05h33m25_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B3/NRCNRCB3-DARK-60090852451_1_488_SE_2016-01-09T09h35m03_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B3/NRCNRCB3-DARK-60091254111_1_488_SE_2016-01-09T14h23m58_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B3/NRCNRCB3-DARK-60091757401_1_488_SE_2016-01-09T18h40m55_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B4/NRCNRCB4-DARK-60090118261_1_489_SE_2016-01-09T02h46m53_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B4/NRCNRCB4-DARK-60090513431_1_489_SE_2016-01-09T05h57m50_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B4/NRCNRCB4-DARK-60090914351_1_489_SE_2016-01-09T09h52m02_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B4/NRCNRCB4-DARK-60091316411_1_489_SE_2016-01-09T14h23m38_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B4/NRCNRCB4-DARK-60091822061_1_489_SE_2016-01-09T18h53m02_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B5/NRCNRCBLONG-DARK-60090141241_1_490_SE_2016-01-09T02h46m50_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B5/NRCNRCBLONG-DARK-60090535381_1_490_SE_2016-01-09T06h17m51_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B5/NRCNRCBLONG-DARK-60090939281_1_490_SE_2016-01-09T10h22m25_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B5/NRCNRCBLONG-DARK-60091338491_1_490_SE_2016-01-09T15h46m43_level1b_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/raw/B5/NRCNRCBLONG-DARK-60101408431_1_490_SE_2016-01-10T15h01m09_level1b_uncal.fits.gz']
NIRCAM_LINEARIZED_DARK_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A1/Linearized_Dark_and_SBRefpix_NRCNRCA1-DARK-60082202011_1_481_SE_2016-01-09T00h03m58_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A1/Linearized_Dark_and_SBRefpix_NRCNRCA1-DARK-60090213141_1_481_SE_2016-01-09T02h53m12_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A1/Linearized_Dark_and_SBRefpix_NRCNRCA1-DARK-60090604481_1_481_SE_2016-01-09T06h52m47_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A1/Linearized_Dark_and_SBRefpix_NRCNRCA1-DARK-60091005411_1_481_SE_2016-01-09T10h56m36_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A1/Linearized_Dark_and_SBRefpix_NRCNRCA1-DARK-60091434481_1_481_SE_2016-01-09T15h50m45_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A2/Linearized_Dark_and_SBRefpix_NRCNRCA2-DARK-60082224241_1_482_SE_2016-01-09T00h10m36_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A2/Linearized_Dark_and_SBRefpix_NRCNRCA2-DARK-60090235001_1_482_SE_2016-01-09T04h17m03_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A2/Linearized_Dark_and_SBRefpix_NRCNRCA2-DARK-60090635511_1_482_SE_2016-01-09T07h05m19_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A2/Linearized_Dark_and_SBRefpix_NRCNRCA2-DARK-60091030561_1_482_SE_2016-01-09T11h03m17_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A2/Linearized_Dark_and_SBRefpix_NRCNRCA2-DARK-60091457131_1_482_SE_2016-01-09T15h50m45_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A3/Linearized_Dark_and_SBRefpix_NRCNRCA3-DARK-60082245481_1_483_SE_2016-01-09T00h04m26_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A3/Linearized_Dark_and_SBRefpix_NRCNRCA3-DARK-60090321241_1_483_SE_2016-01-09T04h17m10_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A3/Linearized_Dark_and_SBRefpix_NRCNRCA3-DARK-60090656591_1_483_SE_2016-01-09T07h31m27_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A3/Linearized_Dark_and_SBRefpix_NRCNRCA3-DARK-60091052561_1_483_SE_2016-01-09T11h28m06_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A3/Linearized_Dark_and_SBRefpix_NRCNRCA3-DARK-60091522581_1_483_SE_2016-01-09T16h30m34_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A4/Linearized_Dark_and_SBRefpix_NRCNRCA4-DARK-60082307391_1_484_SE_2016-01-09T00h04m08_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A4/Linearized_Dark_and_SBRefpix_NRCNRCA4-DARK-60090259591_1_484_SE_2016-01-09T04h16m50_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A4/Linearized_Dark_and_SBRefpix_NRCNRCA4-DARK-60090720041_1_484_SE_2016-01-09T07h58m26_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A4/Linearized_Dark_and_SBRefpix_NRCNRCA4-DARK-60091117441_1_484_SE_2016-01-09T11h52m23_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A4/Linearized_Dark_and_SBRefpix_NRCNRCA4-DARK-60091548131_1_484_SE_2016-01-09T16h30m50_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A5/Linearized_Dark_and_SBRefpix_NRCNRCALONG-DARK-60082329041_1_485_SE_2016-01-09T00h04m16_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A5/Linearized_Dark_and_SBRefpix_NRCNRCALONG-DARK-60090344021_1_485_SE_2016-01-09T04h16m42_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A5/Linearized_Dark_and_SBRefpix_NRCNRCALONG-DARK-60090746381_1_485_SE_2016-01-09T08h21m48_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A5/Linearized_Dark_and_SBRefpix_NRCNRCALONG-DARK-60091140151_1_485_SE_2016-01-09T14h23m49_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/A5/Linearized_Dark_and_SBRefpix_NRCNRCALONG-DARK-60091611271_1_485_SE_2016-01-09T17h16m35_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B1/Linearized_Dark_and_SBRefpix_NRCNRCB1-DARK-60082356471_1_486_SE_2016-01-09T02h47m00_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B1/Linearized_Dark_and_SBRefpix_NRCNRCB1-DARK-60090405201_1_486_SE_2016-01-09T05h33m56_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B1/Linearized_Dark_and_SBRefpix_NRCNRCB1-DARK-60090807581_1_486_SE_2016-01-09T08h48m11_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B1/Linearized_Dark_and_SBRefpix_NRCNRCB1-DARK-60091205311_1_486_SE_2016-01-09T14h30m08_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B1/Linearized_Dark_and_SBRefpix_NRCNRCB1-DARK-60091636021_1_486_SE_2016-01-09T17h16m13_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B2/Linearized_Dark_and_SBRefpix_NRCNRCB2-DARK-60090021181_1_487_SE_2016-01-09T02h51m54_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B2/Linearized_Dark_and_SBRefpix_NRCNRCB2-DARK-60090427541_1_487_SE_2016-01-09T05h33m14_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B2/Linearized_Dark_and_SBRefpix_NRCNRCB2-DARK-60090830131_1_487_SE_2016-01-09T08h59m50_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B2/Linearized_Dark_and_SBRefpix_NRCNRCB2-DARK-60091230011_1_487_SE_2016-01-09T14h23m47_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B2/Linearized_Dark_and_SBRefpix_NRCNRCB2-DARK-60091735131_1_487_SE_2016-01-09T18h09m45_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B3/Linearized_Dark_and_SBRefpix_NRCNRCB3-DARK-60090043151_1_488_SE_2016-01-09T02h53m21_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B3/Linearized_Dark_and_SBRefpix_NRCNRCB3-DARK-60090451471_1_488_SE_2016-01-09T05h33m25_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B3/Linearized_Dark_and_SBRefpix_NRCNRCB3-DARK-60090852451_1_488_SE_2016-01-09T09h35m03_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B3/Linearized_Dark_and_SBRefpix_NRCNRCB3-DARK-60091254111_1_488_SE_2016-01-09T14h23m58_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B3/Linearized_Dark_and_SBRefpix_NRCNRCB3-DARK-60091757401_1_488_SE_2016-01-09T18h40m55_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B4/Linearized_Dark_and_SBRefpix_NRCNRCB4-DARK-60090118261_1_489_SE_2016-01-09T02h46m53_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B4/Linearized_Dark_and_SBRefpix_NRCNRCB4-DARK-60090513431_1_489_SE_2016-01-09T05h57m50_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B4/Linearized_Dark_and_SBRefpix_NRCNRCB4-DARK-60090914351_1_489_SE_2016-01-09T09h52m02_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B4/Linearized_Dark_and_SBRefpix_NRCNRCB4-DARK-60091316411_1_489_SE_2016-01-09T14h23m38_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B4/Linearized_Dark_and_SBRefpix_NRCNRCB4-DARK-60091822061_1_489_SE_2016-01-09T18h53m02_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B5/Linearized_Dark_and_SBRefpix_NRCNRCBLONG-DARK-60090141241_1_490_SE_2016-01-09T02h46m50_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B5/Linearized_Dark_and_SBRefpix_NRCNRCBLONG-DARK-60090535381_1_490_SE_2016-01-09T06h17m51_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B5/Linearized_Dark_and_SBRefpix_NRCNRCBLONG-DARK-60090939281_1_490_SE_2016-01-09T10h22m25_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B5/Linearized_Dark_and_SBRefpix_NRCNRCBLONG-DARK-60091338491_1_490_SE_2016-01-09T15h46m43_uncal.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/darks/linearized/B5/Linearized_Dark_and_SBRefpix_NRCNRCBLONG-DARK-60101408431_1_490_SE_2016-01-10T15h01m09_uncal.fits.gz']

NIRISS_RAW_DARK_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_11_496_SE_2015-12-11T16h05m20_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_12_496_SE_2015-12-11T16h23m51_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_13_496_SE_2015-12-11T16h42m52_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_14_496_SE_2015-12-11T17h01m50_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_15_496_SE_2015-12-11T17h20m40_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_16_496_SE_2015-12-11T17h40m30_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_17_496_SE_2015-12-11T17h59m52_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_18_496_SE_2015-12-11T18h16m31_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_19_496_SE_2015-12-11T18h36m32_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-153451235_20_496_SE_2015-12-11T18h53m52_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_13_496_SE_2017-09-07T04h48m22_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_14_496_SE_2017-09-07T05h06m42_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_15_496_SE_2017-09-07T05h28m22_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_16_496_SE_2017-09-07T05h47m42_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_17_496_SE_2017-09-07T06h09m02_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_18_496_SE_2017-09-07T06h29m12_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_19_496_SE_2017-09-07T06h49m52_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_20_496_SE_2017-09-07T07h09m22_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_21_496_SE_2017-09-07T07h29m52_dms_uncal.fits.gz',
                        'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/raw/NISNIRISSDARK-172500017_22_496_SE_2017-09-07T07h50m32_dms_uncal.fits.gz']
NIRISS_LINEARIZED_DARK_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_11_496_SE_2015-12-11T16h05m20_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_12_496_SE_2015-12-11T16h23m51_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_13_496_SE_2015-12-11T16h42m52_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_14_496_SE_2015-12-11T17h01m50_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_15_496_SE_2015-12-11T17h20m40_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_16_496_SE_2015-12-11T17h40m30_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_17_496_SE_2015-12-11T17h59m52_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_18_496_SE_2015-12-11T18h16m31_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_19_496_SE_2015-12-11T18h36m32_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-153451235_20_496_SE_2015-12-11T18h53m52_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_13_496_SE_2017-09-07T04h48m22_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_14_496_SE_2017-09-07T05h06m42_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_15_496_SE_2017-09-07T05h28m22_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_16_496_SE_2017-09-07T05h47m42_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_17_496_SE_2017-09-07T06h09m02_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_18_496_SE_2017-09-07T06h29m12_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_19_496_SE_2017-09-07T06h49m52_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_20_496_SE_2017-09-07T07h09m22_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_21_496_SE_2017-09-07T07h29m52_dms_uncal_linear_dark_prep_object.fits.gz',
                               'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/niriss/darks/linearized/Linearized_Dark_and_SBRefpix_NISNIRISSDARK-172500017_22_496_SE_2017-09-07T07h50m32_dms_uncal_linear_dark_prep_object.fits.gz']

FGS_RAW_DARK_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/29722_1x88_FGSF03512-D-NR-G2-5339214947_1_498_SE_2015-12-05T22h27m19_dms_uncal.fits.gz',
                     'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/29782_1x88_FGSF03872-PAR-5340074326_1_498_SE_2015-12-06T12h22m47_dms_uncal.fits.gz',
                     'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/29813_1x88_FGSF037221-MR-2-5340161743_1_498_SE_2015-12-06T16h45m10_dms_uncal.fits.gz',
                     'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/30632_1x88_FGSF03511-D-NR-G1-5346180117_1_497_SE_2015-12-12T19h00m12_dms_uncal.fits.gz',
                     'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/30670_1x88_FGSF03511-D-NR-G2-5346181816_1_498_SE_2015-12-12T21h31m01_dms_uncal.fits.gz',
                     'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/30742_1x88_FGSF03871-PAR-5347035139_1_498_SE_2015-12-13T05h23m30_dms_uncal.fits.gz',
                     'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/30749_1x88_FGSF03881-PAR-5347043800_1_497_SE_2015-12-13T09h02m01_dms_uncal.fits.gz',
                     'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/raw/30829_1x88_FGSF037111-G1NRNC-5347151640_1_497_SE_2015-12-13T16h28m38_dms_uncal.fits.gz']
FGS_LINEARIZED_DARK_URLS = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/29722_1x88_FGSF03512-D-NR-G2-5339214947_1_498_SE_2015-12-05T22h27m19_dms_uncal_linearized.fits.gz',
                            'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/29782_1x88_FGSF03872-PAR-5340074326_1_498_SE_2015-12-06T12h22m47_dms_uncal_linearized.fits.gz',
                            'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/29813_1x88_FGSF037221-MR-2-5340161743_1_498_SE_2015-12-06T16h45m10_dms_uncal_linearized.fits.gz',
                            'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/30632_1x88_FGSF03511-D-NR-G1-5346180117_1_497_SE_2015-12-12T19h00m12_dms_uncal_linearized.fits.gz',
                            'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/30670_1x88_FGSF03511-D-NR-G2-5346181816_1_498_SE_2015-12-12T21h31m01_dms_uncal_linearized.fits.gz',
                            'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/30742_1x88_FGSF03871-PAR-5347035139_1_498_SE_2015-12-13T05h23m30_dms_uncal_linearized.fits.gz',
                            'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/30749_1x88_FGSF03881-PAR-5347043800_1_497_SE_2015-12-13T09h02m01_dms_uncal_linearized.fits.gz',
                            'https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/fgs/darks/linearized/30829_1x88_FGSF037111-G1NRNC-5347151640_1_497_SE_2015-12-13T16h28m38_dms_uncal_linearized.fits.gz']

TEMP_DISTORTION_REFERENCE_FILES = ['https://data.science.stsci.edu/redirect/JWST/jwst-simulations/mirage_reference_files/nircam/reference_files/nircam_distortion_files.tar.gz']

DISK_USAGE = {'nircam': {'crs': 1.1, 'psfs': 23, 'raw_darks': 79, 'lin_darks': 319, 'grism': 0.005},
              'niriss': {'crs': 0.26, 'psfs': 0.87, 'raw_darks': 31, 'lin_darks': 121, 'grism': 0.167},
              'fgs': {'crs': 0.31, 'psfs': .04, 'raw_darks': 11, 'lin_darks': 39}}

def download_file(url, file_name, output_directory='./'):
    """Download into the current working directory the
    file from Box given the direct URL

    Parameters
    ----------
    url : str
        URL to the file to be downloaded

    Returns
    -------
    download_filename : str
        Name of the downloaded file
    """
    download_filename = os.path.join(output_directory, file_name)

    # Only download the file if it doesn't already exist
    if not os.path.isfile(download_filename):
        print('Downloading: {}'.format(file_name))
        with requests.get(url, stream=True) as response:
            if response.status_code != 200:
                raise RuntimeError("Wrong URL - {}".format(url))
            with open(download_filename, 'wb') as f:
                for chunk in response.iter_content(chunk_size=2048):
                    if chunk:
                        f.write(chunk)
        print('Download of {} complete.'.format(file_name))
    else:
        print('{} already exists. Skipping download.'.format(file_name))
    return download_filename


def download_reffiles(directory, instrument='all', dark_type='linearized',
                      skip_darks=False, skip_cosmic_rays=False, skip_psfs=False,
                      skip_grism=False):
    """Download tarred and gzipped reference files. Expand, unzip and
    organize into the necessary directory structure such that Mirage
    can use them.

    Parameters
    ----------
    directory : str
        Directory into which the reference files are placed. This will
        be the directory set to the MIRAGE_DATA environment variable

    instrument : str
        Instruments for which to download data. Single string with
        comma-separated instrument names.

        If ``all``: download files for NIRCam, NIRISS, and FGS.

        If the name of an individual instrument (e.g. 'nircam'),
        download only the data for that instrument.

        If a list of instrument names, (e.g. 'nircam, niriss') download
        all data for those instruments.

    dark_type : str
        Type of dark current files to download. Options are:
        'linearized': download linearize dark current ramps
        'raw': download raw dark current ramps
        'both': download both raw and linearized dark current ramps

    skip_darks : bool
        If False (default), download the requested darks. If True,
        do not download the darks

    skip_comsic_rays : bool
        If False (default), download the requested cosmic ray libraries.
        If True, do not download the cosmic ray library.

    skip_psfs : bool
        If False (default), download the requested PSF libraries.
        If True, do not download the libraries.

    skip_grism : bool
        If False (default), download the grism-related reference files
        If True, do not download the grism data.
    """
    # Be sure the input instrument is a list
    file_list = get_file_list(instrument.lower(), dark_type.lower(),
                              skip_darks=skip_darks,
                              skip_cosmic_rays=skip_cosmic_rays,
                              skip_psfs=skip_psfs, skip_grism=skip_grism)

    # Download everything first
    for file_url in file_list:
        filename = os.path.split(file_url)[-1]
        local_file = os.path.join(directory, filename)
        download_file(file_url, filename, directory)

    # Now untar/organize. This way if the download is interrupted, it can
    # pick up where it left off, since no downloaded files will have been
    # moved yet.
    for file_url in file_list:
        filename = os.path.split(file_url)[-1]
        local_file = os.path.join(directory, filename)
        # Unzip and untar file
        if 'tar.gz' in local_file:
            print('Unzipping/extracting {}'.format(filename))
            file_object = tarfile.open(name=local_file, mode='r:gz')
            file_object.extractall(path=directory)
        else:
            # Darks need to be unzipped into the correct directory
            if 'linearized' in filename.lower():
                cal = 'linearized'
            else:
                cal = 'raw'

            # Determine directory
            if 'NRCNRC' in filename:
                det_str = filename.split('NRCNRC')[1].split('-')[0]
                if 'LONG' in det_str:
                    det_str.replace('LONG', '5')
                darks_dir = os.path.join(directory, 'mirage_data', 'nircam', 'darks')
                sub_directory = os.path.join(darks_dir, cal, det_str)
            elif 'NIRISS' in filename.lower():
                darks_dir = os.path.join(directory, 'mirage_data', 'niriss', 'darks')
                sub_directory = os.path.join(darks_dir, cal)
                #sub_directory = os.path.join(directory, 'mirage_data', 'niriss', 'darks', cal)
            elif 'FGS' in filename:
                darks_dir = os.path.join(directory, 'mirage_data', 'fgs', 'darks')
                sub_directory = os.path.join(darks_dir, cal)

            # Create the directory if it does not yet exist
            ensure_dir_exists(darks_dir)
            ensure_dir_exists(os.path.join(darks_dir, cal))
            ensure_dir_exists(sub_directory)

            final_location = os.path.join(sub_directory, filename)

            # Move the zipped file into the correct subdirectory
            if not os.path.isfile(final_location):
                print('Moving {} to {}'.format(filename, sub_directory))
                shutil.move(local_file, final_location)

            # Unzip
            unzipped_filename = final_location.replace('.gz', '')
            if not os.path.isfile(unzipped_filename):
                unzip_file(final_location)
            else:
                print('Unzipped file {} already exists. Skipping unzip.'.format(unzipped_filename))

    full_dir = os.path.abspath(directory)
    print(('Mirage reference files downloaded and extracted. \nBefore '
           'using Mirage, be sure to set the MIRAGE_DATA environment '
           'variable to point to:\n{}/mirage_data'.format(full_dir)))
    print('In bash: ')
    print('export MIRAGE_DATA="{}"'.format(os.path.join(full_dir, 'mirage_data')))


def get_file_list(instruments, dark_current, skip_darks=False, skip_cosmic_rays=False,
                  skip_psfs=False, skip_grism=False):
    """Collect the list of URLs corresponding to the Mirage reference
    files to be downloaded

    Parameters
    ----------
    instruments : list
        List of instrument names for which to download data

    dark_current : str
        Type of dark current files to download. Options are:
        'linearized': download linearize dark current ramps
        'raw': download raw dark current ramps
        'both': download both raw and linearized dark current ramps

    skip_darks : bool
        If False (default), include the requested darks. If True,
        do not include the darks

    skip_comsic_rays : bool
        If False (default), include the requested cosmic ray libraries.
        If True, do not include the cosmic ray library.

    skip_psfs : bool
        If False (default), include the requested PSF libraries.
        If True, do not include the libraries.

    skip_grism : bool
        If False (default), include the grism-related reference files.
        If True, do not include the grism files.

    Returns
    -------
    urls : list
        List of tuples. Each tuple contains:
        (URL for downloading, name of file)
    """
    urls = []
    total_download_size = 0.
    instrument_names = [name.strip().lower() for name in instruments.split(',')]

    if 'all' in instrument_names:
        instrument_names = ['nircam', 'niriss', 'fgs']

    for instrument_name in instrument_names:
        # NIRCam
        if instrument_name.lower() == 'nircam':

            if not skip_cosmic_rays:
                urls.extend(NIRCAM_CR_LIBRARY_URL)
                added_size = DISK_USAGE['nircam']['crs']
                total_download_size += added_size
                print('Size of NIRCam cosmic ray library file: {} Gb'.format(added_size))

            if not skip_psfs:
                urls.extend(NIRCAM_GRIDDED_PSF_URLS)
                added_size = DISK_USAGE['nircam']['psfs']
                total_download_size += added_size
                print('Size of NIRCam PSF library file: {} Gb'.format(added_size))

            if not skip_darks:
                if dark_current in ['linearized', 'both']:
                    urls.extend(NIRCAM_LINEARIZED_DARK_URLS)
                    added_size = DISK_USAGE['nircam']['lin_darks']
                    total_download_size += added_size
                    print('Size of NIRCam linearized dark files: {} Gb'.format(added_size))
                elif dark_current in ['raw', 'both']:
                    urls.extend(NIRCAM_RAW_DARK_URLS)
                    added_size = DISK_USAGE['nircam']['raw_darks']
                    total_download_size += added_size
                    print('Size of NIRCam raw dark files: {} Gb'.format(added_size))

            if not skip_grism:
                # Get grism-related files
                urls.extend(NIRCAM_GRISM_FILES)
                added_size = DISK_USAGE['nircam']['grism']
                total_download_size += added_size
                print('Size of NIRCam grism files: {} Gb'.format(added_size))

            # Get the temporary distortion reference files with the
            # correct coefficients
            urls.extend(TEMP_DISTORTION_REFERENCE_FILES)

        # NIRISS
        elif instrument_name.lower() == 'niriss':
            if not skip_cosmic_rays:
                urls.extend(NIRISS_CR_LIBRARY_URL)
                added_size = DISK_USAGE['niriss']['crs']
                total_download_size += added_size
                print('Size of NIRISS cosmic ray library file: {} Gb'.format(added_size))

            if not skip_psfs:
                urls.extend(NIRISS_GRIDDED_PSF_URLS)
                added_size = DISK_USAGE['niriss']['psfs']
                total_download_size += added_size
                print('Size of NIRISS PSF library file: {} Gb'.format(added_size))

            if not skip_darks:
                if dark_current in ['linearized', 'both']:
                    urls.extend(NIRISS_LINEARIZED_DARK_URLS)
                    added_size = DISK_USAGE['niriss']['lin_darks']
                    total_download_size += added_size
                    print('Size of NIRISS linearized dark files: {} Gb'.format(added_size))
                elif dark_current in ['raw', 'both']:
                    urls.extend(NIRISS_RAW_DARK_URLS)
                    added_size = DISK_USAGE['niriss']['raw_darks']
                    total_download_size += added_size
                    print('Size of NIRISS raw dark files: {} Gb'.format(added_size))

            if not skip_grism:
                # Get grism-related files
                urls.extend(NIRISS_GRISM_FILES)
                added_size = DISK_USAGE['niriss']['grism']
                total_download_size += added_size
                print('Size of NIRISS grism files: {} Gb'.format(added_size))

        # FGS
        elif instrument_name.lower() == 'fgs':
            if not skip_cosmic_rays:
                urls.extend(FGS_CR_LIBRARY_URL)
                added_size = DISK_USAGE['fgs']['crs']
                total_download_size += added_size
                print('Size of FGS cosmic ray library file: {} Gb'.format(added_size))

            if not skip_psfs:
                urls.extend(FGS_GRIDDED_PSF_URLS)
                added_size = DISK_USAGE['fgs']['psfs']
                total_download_size += added_size
                print('Size of FGS PSF library file: {} Gb'.format(added_size))

            if not skip_darks:
                if dark_current in ['linearized', 'both']:
                    urls.extend(FGS_LINEARIZED_DARK_URLS)
                    added_size = DISK_USAGE['fgs']['lin_darks']
                    total_download_size += added_size
                    print('Size of FGS linearized dark files: {} Gb'.format(added_size))
                elif dark_current in ['raw', 'both']:
                    urls.extend(FGS_RAW_DARK_URLS)
                    added_size = DISK_USAGE['fgs']['raw_darks']
                    total_download_size += added_size
                    print('Size of FGS raw dark files: {} Gb'.format(added_size))

    print("Total size of files to be downloaded: {} Gb".format(total_download_size))
    return urls


def unzip_file(filename):
    """Unzip a file

    Parameters
    ----------
    filename : str
        Name of file to unzip

    dir_name : str
        Directory into which the file is unzipped
    """
    if not os.path.isfile(filename):
        print('File {} does not exist.'.format(filename))
    print('Unzipping {}'.format(filename))
    with gzip.open(filename, 'rb') as file_in:
        unzipped_name = filename.replace('.gz', '')
        with open(unzipped_name, 'wb') as file_out:
            shutil.copyfileobj(file_in, file_out)


#if __name__ == '__main__':
#      params = parse_args()
#
#      download_reffiles(directory, instrument=instrument, dark_type=dark_type,
#                        skip_darks=skip_darks, skip_cosmic_rays=skip_cosmic_rays,
#                        skip_psfs=skip_psfs)
